__all__ = [
    "EXTRACTORS",
    "HF_EXTRACTORS",
]

base_import: str = "wisent.core.contrastive_pairs.huggingface_pairs.hf_task_extractors."

EXTRACTORS: dict[str, str] = {
    # Math benchmarks
    "aime": f"{base_import}aime:AIMEExtractor",
    "aime2024": f"{base_import}aime2024:AIME2024Extractor",
    "aime2025": f"{base_import}aime2025:AIME2025Extractor",
    "asdiv_cot_llama": f"{base_import}math500:MATH500Extractor",
    "chain_of_thought": f"{base_import}math500:MATH500Extractor",
    "gsm8k": f"{base_import}gsm8k_extractor:GSM8KExtractor",
    "gsm8k_cot": f"{base_import}math500:MATH500Extractor",
    "gsm8k_cot_llama": f"{base_import}math500:MATH500Extractor",
    "gsm8k_cot_self_consistency": f"{base_import}math500:MATH500Extractor",
    "gsm8k_llama": f"{base_import}math500:MATH500Extractor",
    "gsm8k_platinum_cot": f"{base_import}math500:MATH500Extractor",
    "gsm8k_platinum_cot_llama": f"{base_import}math500:MATH500Extractor",
    "gsm8k_platinum_cot_self_consistency": f"{base_import}math500:MATH500Extractor",
    "hmmt": f"{base_import}hmmt:HMMTExtractor",
    "hmmt_feb_2025": f"{base_import}hmmt:HMMTExtractor",
    "livemathbench": f"{base_import}livemathbench:LiveMathBenchExtractor",
    # v202412 - December 2024 release
    "livemathbench_cnmo_en": f"{base_import}livemathbench:LiveMathBenchCnmoEnExtractor",
    "livemathbench_cnmo_cn": f"{base_import}livemathbench:LiveMathBenchCnmoCnExtractor",
    "livemathbench_ccee_en": f"{base_import}livemathbench:LiveMathBenchCceeEnExtractor",
    "livemathbench_ccee_cn": f"{base_import}livemathbench:LiveMathBenchCceeCnExtractor",
    "livemathbench_amc_en": f"{base_import}livemathbench:LiveMathBenchAmcEnExtractor",
    "livemathbench_amc_cn": f"{base_import}livemathbench:LiveMathBenchAmcCnExtractor",
    "livemathbench_wlpmc_en": f"{base_import}livemathbench:LiveMathBenchWlpmcEnExtractor",
    "livemathbench_wlpmc_cn": f"{base_import}livemathbench:LiveMathBenchWlpmcCnExtractor",
    "livemathbench_hard_en": f"{base_import}livemathbench:LiveMathBenchHardEnExtractor",
    "livemathbench_hard_cn": f"{base_import}livemathbench:LiveMathBenchHardCnExtractor",
    # v202505 - May 2025 release
    "livemathbench_v202505_all_en": f"{base_import}livemathbench:LiveMathBenchV202505AllEnExtractor",
    "livemathbench_v202505_hard_en": f"{base_import}livemathbench:LiveMathBenchV202505HardEnExtractor",
    "math": f"{base_import}math:MATHExtractor",
    "math_500": f"{base_import}math500:MATH500Extractor",
    "math500": f"{base_import}math500:MATH500Extractor",
    "polymath": f"{base_import}polymath:PolyMathExtractor",
    # Arabic
    "polymath_ar_top": f"{base_import}polymath:PolyMathARTopExtractor",
    "polymath_ar_high": f"{base_import}polymath:PolyMathARHighExtractor",
    "polymath_ar_medium": f"{base_import}polymath:PolyMathARMediumExtractor",
    "polymath_ar_low": f"{base_import}polymath:PolyMathARLowExtractor",
    # Bengali
    "polymath_bn_top": f"{base_import}polymath:PolyMathBNTopExtractor",
    "polymath_bn_high": f"{base_import}polymath:PolyMathBNHighExtractor",
    "polymath_bn_medium": f"{base_import}polymath:PolyMathBNMediumExtractor",
    "polymath_bn_low": f"{base_import}polymath:PolyMathBNLowExtractor",
    # German
    "polymath_de_top": f"{base_import}polymath:PolyMathDETopExtractor",
    "polymath_de_high": f"{base_import}polymath:PolyMathDEHighExtractor",
    "polymath_de_medium": f"{base_import}polymath:PolyMathDEMediumExtractor",
    "polymath_de_low": f"{base_import}polymath:PolyMathDELowExtractor",
    # English
    "polymath_en_top": f"{base_import}polymath:PolyMathENTopExtractor",
    "polymath_en_high": f"{base_import}polymath:PolyMathENHighExtractor",
    "polymath_en_medium": f"{base_import}polymath:PolyMathENMediumExtractor",
    "polymath_en_low": f"{base_import}polymath:PolyMathENLowExtractor",
    # Spanish
    "polymath_es_top": f"{base_import}polymath:PolyMathESTopExtractor",
    "polymath_es_high": f"{base_import}polymath:PolyMathESHighExtractor",
    "polymath_es_medium": f"{base_import}polymath:PolyMathESMediumExtractor",
    "polymath_es_low": f"{base_import}polymath:PolyMathESLowExtractor",
    # French
    "polymath_fr_top": f"{base_import}polymath:PolyMathFRTopExtractor",
    "polymath_fr_high": f"{base_import}polymath:PolyMathFRHighExtractor",
    "polymath_fr_medium": f"{base_import}polymath:PolyMathFRMediumExtractor",
    "polymath_fr_low": f"{base_import}polymath:PolyMathFRLowExtractor",
    # Indonesian
    "polymath_id_top": f"{base_import}polymath:PolyMathIDTopExtractor",
    "polymath_id_high": f"{base_import}polymath:PolyMathIDHighExtractor",
    "polymath_id_medium": f"{base_import}polymath:PolyMathIDMediumExtractor",
    "polymath_id_low": f"{base_import}polymath:PolyMathIDLowExtractor",
    # Italian
    "polymath_it_top": f"{base_import}polymath:PolyMathITTopExtractor",
    "polymath_it_high": f"{base_import}polymath:PolyMathITHighExtractor",
    "polymath_it_medium": f"{base_import}polymath:PolyMathITMediumExtractor",
    "polymath_it_low": f"{base_import}polymath:PolyMathITLowExtractor",
    # Japanese
    "polymath_ja_top": f"{base_import}polymath:PolyMathJATopExtractor",
    "polymath_ja_high": f"{base_import}polymath:PolyMathJAHighExtractor",
    "polymath_ja_medium": f"{base_import}polymath:PolyMathJAMediumExtractor",
    "polymath_ja_low": f"{base_import}polymath:PolyMathJALowExtractor",
    # Korean
    "polymath_ko_top": f"{base_import}polymath:PolyMathKOTopExtractor",
    "polymath_ko_high": f"{base_import}polymath:PolyMathKOHighExtractor",
    "polymath_ko_medium": f"{base_import}polymath:PolyMathKOMediumExtractor",
    "polymath_ko_low": f"{base_import}polymath:PolyMathKOLowExtractor",
    # Malay
    "polymath_ms_top": f"{base_import}polymath:PolyMathMSTopExtractor",
    "polymath_ms_high": f"{base_import}polymath:PolyMathMSHighExtractor",
    "polymath_ms_medium": f"{base_import}polymath:PolyMathMSMediumExtractor",
    "polymath_ms_low": f"{base_import}polymath:PolyMathMSLowExtractor",
    # Portuguese
    "polymath_pt_top": f"{base_import}polymath:PolyMathPTTopExtractor",
    "polymath_pt_high": f"{base_import}polymath:PolyMathPTHighExtractor",
    "polymath_pt_medium": f"{base_import}polymath:PolyMathPTMediumExtractor",
    "polymath_pt_low": f"{base_import}polymath:PolyMathPTLowExtractor",
    # Russian
    "polymath_ru_top": f"{base_import}polymath:PolyMathRUTopExtractor",
    "polymath_ru_high": f"{base_import}polymath:PolyMathRUHighExtractor",
    "polymath_ru_medium": f"{base_import}polymath:PolyMathRUMediumExtractor",
    "polymath_ru_low": f"{base_import}polymath:PolyMathRULowExtractor",
    # Swahili
    "polymath_sw_top": f"{base_import}polymath:PolyMathSWTopExtractor",
    "polymath_sw_high": f"{base_import}polymath:PolyMathSWHighExtractor",
    "polymath_sw_medium": f"{base_import}polymath:PolyMathSWMediumExtractor",
    "polymath_sw_low": f"{base_import}polymath:PolyMathSWLowExtractor",
    # Telugu
    "polymath_te_top": f"{base_import}polymath:PolyMathTETopExtractor",
    "polymath_te_high": f"{base_import}polymath:PolyMathTEHighExtractor",
    "polymath_te_medium": f"{base_import}polymath:PolyMathTEMediumExtractor",
    "polymath_te_low": f"{base_import}polymath:PolyMathTELowExtractor",
    # Thai
    "polymath_th_top": f"{base_import}polymath:PolyMathTHTopExtractor",
    "polymath_th_high": f"{base_import}polymath:PolyMathTHHighExtractor",
    "polymath_th_medium": f"{base_import}polymath:PolyMathTHMediumExtractor",
    "polymath_th_low": f"{base_import}polymath:PolyMathTHLowExtractor",
    # Vietnamese
    "polymath_vi_top": f"{base_import}polymath:PolyMathVITopExtractor",
    "polymath_vi_high": f"{base_import}polymath:PolyMathVIHighExtractor",
    "polymath_vi_medium": f"{base_import}polymath:PolyMathVIMediumExtractor",
    "polymath_vi_low": f"{base_import}polymath:PolyMathVILowExtractor",
    # Chinese
    "polymath_zh_top": f"{base_import}polymath:PolyMathZHTopExtractor",
    "polymath_zh_high": f"{base_import}polymath:PolyMathZHHighExtractor",
    "polymath_zh_medium": f"{base_import}polymath:PolyMathZHMediumExtractor",
    "polymath_zh_low": f"{base_import}polymath:PolyMathZHLowExtractor",

    # Coding benchmarks
    "humaneval": f"{base_import}humaneval:HumanEvalExtractor",
    "humaneval_64": f"{base_import}humaneval:HumanEval64Extractor",
    "humaneval_plus": f"{base_import}humaneval:HumanEvalPlusExtractor",
    "humaneval_instruct": f"{base_import}humaneval:HumanEvalInstructExtractor",
    "humaneval_64_instruct": f"{base_import}humaneval:HumanEval64InstructExtractor",
    "humanevalpack": f"{base_import}humanevalpack:HumanevalpackExtractor",
    "apps": f"{base_import}apps:AppsExtractor",
    "conala": f"{base_import}conala:ConalaExtractor",
    "concode": f"{base_import}concode:ConcodeExtractor",
    "ds_1000": f"{base_import}ds_1000:Ds1000Extractor",
    "ds1000": f"{base_import}ds_1000:Ds1000Extractor",
    "mercury": f"{base_import}mercury:MercuryExtractor",
    "recode": f"{base_import}recode:RecodeExtractor",
    "multipl_e": f"{base_import}multipl_e:MultiplEExtractor",
    "multiple_py": f"{base_import}multipl_e:MultiplEExtractor",
    "multiple_js": f"{base_import}multipl_e:MultiplEExtractor",
    "multiple_java": f"{base_import}multipl_e:MultiplEExtractor",
    "multiple_cpp": f"{base_import}multipl_e:MultiplEExtractor",
    "multiple_rs": f"{base_import}multipl_e:MultiplEExtractor",
    "multiple_go": f"{base_import}multipl_e:MultiplEExtractor",
    "codexglue": f"{base_import}codexglue:CodexglueExtractor",
    "livecodebench": f"{base_import}livecodebench:LivecodebenchExtractor",

    # Reasoning benchmarks
    "super_gpqa": f"{base_import}super_gpqa:SuperGpqaExtractor",
    "supergpqa": f"{base_import}super_gpqa:SuperGpqaExtractor",
    "supergpqa_physics": f"{base_import}super_gpqa:SuperGpqaExtractor",
    "supergpqa_chemistry": f"{base_import}super_gpqa:SuperGpqaExtractor",
    "supergpqa_biology": f"{base_import}super_gpqa:SuperGpqaExtractor",
    "hle": f"{base_import}hle:HleExtractor",
    "hle_exact_match": f"{base_import}hle:HleExtractor",
    "hle_multiple_choice": f"{base_import}hle:HleExtractor",

    # Database/Table benchmarks
    "tag": f"{base_import}tag:TagExtractor",
    "Tag": f"{base_import}tag:TagExtractor",

    # Medical benchmarks
    "meddialog": f"{base_import}meddialog:MeddialogExtractor",
    "meddialog_qsumm": f"{base_import}meddialog:MeddialogExtractor",
    "meddialog_qsumm_perplexity": f"{base_import}meddialog:MeddialogExtractor",
    "meddialog_raw_dialogues": f"{base_import}meddialog:MeddialogExtractor",
    "meddialog_raw_perplexity": f"{base_import}meddialog:MeddialogExtractor",

    # MMLU-SR benchmarks (all variants use the same extractor)
    "mmlusr": f"{base_import}mmlusr:MMLUSRExtractor",
    "mmlusr_answer_only": f"{base_import}mmlusr:MMLUSRExtractor",
    "mmlusr_question_only": f"{base_import}mmlusr:MMLUSRExtractor",
    "mmlusr_question_and_answer": f"{base_import}mmlusr:MMLUSRExtractor",

    # Newly moved from lm_eval_pairs
    "atis": f"{base_import}atis:AtisExtractor",
    "babilong": f"{base_import}babilong:BabilongExtractor",
    "bangla_mmlu": f"{base_import}bangla_mmlu:BanglaMmluExtractor",
    # Winogender - gender bias benchmark (HuggingFace oskarvanderwal/winogender)
    "winogender": f"{base_import}winogender:WinogenderHfExtractor",
    "basqueglue": f"{base_import}basqueglue:BasqueglueExtractor",
    "bec2016eu": f"{base_import}bec2016eu:Bec2016euExtractor",
    "boolq_seq2seq": f"{base_import}boolq_seq2seq:BoolqSeq2seqExtractor",
    "doc_vqa": f"{base_import}doc_vqa:DocVQAExtractor",
    "ds1000": f"{base_import}ds1000:Ds1000Extractor",
    "evalita_mp": f"{base_import}evalita_mp:EvalitaMpExtractor",
    "flores": f"{base_import}flores:FloresExtractor",
    "humanevalpack": f"{base_import}humanevalpack:HumanevalpackExtractor",
    "iwslt2017_ar_en": f"{base_import}iwslt2017_ar_en:Iwslt2017ArEnExtractor",
    "iwslt2017_en_ar": f"{base_import}iwslt2017_en_ar:Iwslt2017EnArExtractor",
    "llama": f"{base_import}llama:LlamaExtractor",
    "multimedqa": f"{base_import}multimedqa:MultimedqaExtractor",
    "multiple": f"{base_import}multiple:MultipleExtractor",
    "openllm": f"{base_import}openllm:OpenllmExtractor",
    "pythia": f"{base_import}pythia:PythiaExtractor",
    "squad2": f"{base_import}squad2:SQuAD2Extractor",
    "stsb": f"{base_import}stsb:StsbExtractor",
    "super_glue_lm_eval_v1": f"{base_import}super_glue_lm_eval_v1:SuperGlueLmEvalV1Extractor",
    "super_glue_lm_eval_v1_seq2seq": f"{base_import}super_glue_lm_eval_v1_seq2seq:SuperGlueLmEvalV1Seq2seqExtractor",
    "super_glue_t5_prompt": f"{base_import}super_glue_t5_prompt:SuperGlueT5PromptExtractor",
    "tmlu": f"{base_import}tmlu:TmluExtractor",
    "wiceu": f"{base_import}wiceu:WiceuExtractor",
    "wmt_ro_en_t5_prompt": f"{base_import}wmt_ro_en_t5_prompt:WmtRoEnT5PromptExtractor",

    # Newly created extractors
    # Note: unitxt tasks (xsum, cnn_dailymail, dbpedia_14, ethos_binary, etc.) use lm-eval unitxt extractor
    # Note: babi removed - use lm-eval version instead (HF version has loading issues)
    "bhtc_v2": f"{base_import}bhtc_v2:BhtcV2Extractor",
    "drop": f"{base_import}squad_extractor:DROPExtractor",
    "sciq": f"{base_import}babi_extractor:SciQExtractor",
    "squadv2": f"{base_import}squad_extractor:SQuADv2Extractor",
    "basque-glue": f"{base_import}basqueglue:BasqueglueExtractor",
    "evalita-sp_sum_task_fp-small_p1": f"{base_import}evalita_sp_sum_task_fp_small_p1:EvalitaSpSumTaskFpSmallP1Extractor",
    "flan_held_in": f"{base_import}flan_held_in:FlanHeldInExtractor",
    "gpt3_translation_benchmarks": f"{base_import}gpt3_translation_benchmarks:Gpt3TranslationBenchmarksExtractor",
    "multiple_choice": f"{base_import}multiple_choice:MultipleChoiceExtractor",
    "penn_treebank": f"{base_import}penn_treebank:PennTreebankExtractor",
    "ptb": f"{base_import}penn_treebank:PennTreebankExtractor",
    "self_consistency": f"{base_import}self_consistency:SelfConsistencyExtractor",
    "t0_eval": f"{base_import}t0_eval:T0EvalExtractor",
    "vaxx_stance": f"{base_import}vaxx_stance:VaxxStanceExtractor",
    "wikitext103": f"{base_import}wikitext103:Wikitext103Extractor",

    # NOTE: truthfulqa_gen/truthfulqa_generation now use lm-eval extractor (see group_task_manifests/truthfulqa.py)

    # Factuality benchmarks (NOT lm-eval)
    "simpleqa": f"{base_import}simpleqa:SimpleQAExtractor",
    "simple_qa": f"{base_import}simpleqa:SimpleQAExtractor",

    # MMLU variants (NOT lm-eval)
    "mmlu_redux": f"{base_import}mmlu_redux:MMLUReduxExtractor",
    "mmlu-redux": f"{base_import}mmlu_redux:MMLUReduxExtractor",

    # Multi-hop reasoning benchmarks (NOT lm-eval)
    "frames": f"{base_import}frames:FRAMESExtractor",
    "frames_benchmark": f"{base_import}frames:FRAMESExtractor",

    # Medical/Health benchmarks (NOT lm-eval)
    "healthbench": f"{base_import}healthbench:HealthBenchExtractor",
    "health_bench": f"{base_import}healthbench:HealthBenchExtractor",

    # Software Engineering benchmarks (NOT lm-eval)
    "swe_bench_verified": f"{base_import}swe_bench_verified:SWEBenchVerifiedExtractor",
    "swe-bench-verified": f"{base_import}swe_bench_verified:SWEBenchVerifiedExtractor",
    "swebench_verified": f"{base_import}swe_bench_verified:SWEBenchVerifiedExtractor",

    # Scientific Computing benchmarks (NOT lm-eval)
    "scicode": f"{base_import}scicode:SciCodeExtractor",
    "sci_code": f"{base_import}scicode:SciCodeExtractor",

    # Instruction Following benchmarks (NOT lm-eval)
    "alpaca_eval": f"{base_import}alpaca_eval:AlpacaEvalExtractor",
    "alpacaeval": f"{base_import}alpaca_eval:AlpacaEvalExtractor",
    "alpaca_eval_2": f"{base_import}alpaca_eval:AlpacaEvalExtractor",

    # Arena/Chat benchmarks (NOT lm-eval)
    "arena_hard": f"{base_import}arena_hard:ArenaHardExtractor",
    "arena-hard": f"{base_import}arena_hard:ArenaHardExtractor",
    "arenahard": f"{base_import}arena_hard:ArenaHardExtractor",

    # Safety/Refusal benchmarks (NOT lm-eval)
    "sorry_bench": f"{base_import}sorry_bench:SorryBenchExtractor",
    "sorry-bench": f"{base_import}sorry_bench:SorryBenchExtractor",
    "sorrybench": f"{base_import}sorry_bench:SorryBenchExtractor",
    "jailbreakbench": f"{base_import}jailbreakbench:JailbreakBenchExtractor",
    "jailbreak_bench": f"{base_import}jailbreakbench:JailbreakBenchExtractor",
    "jbb": f"{base_import}jailbreakbench:JailbreakBenchExtractor",
    "harmbench": f"{base_import}harmbench:HarmBenchExtractor",
    "harm_bench": f"{base_import}harmbench:HarmBenchExtractor",
    "harm-bench": f"{base_import}harmbench:HarmBenchExtractor",

    # Over-refusal benchmarks (NOT lm-eval)
    "or_bench": f"{base_import}or_bench:ORBenchExtractor",
    "or-bench": f"{base_import}or_bench:ORBenchExtractor",
    "orbench": f"{base_import}or_bench:ORBenchExtractor",
    "or_bench_hard": f"{base_import}or_bench:ORBenchExtractor",
    "or_bench_80k": f"{base_import}or_bench:ORBenchExtractor",

    # Function calling benchmarks (NOT lm-eval)
    "bfcl": f"{base_import}bfcl:BFCLExtractor",
    "berkeley_function_calling": f"{base_import}bfcl:BFCLExtractor",
    "function_calling_leaderboard": f"{base_import}bfcl:BFCLExtractor",

    # Hallucination benchmarks (NOT lm-eval)
    "halueval": f"{base_import}halueval:HaluEvalExtractor",
    "halu_eval": f"{base_import}halueval:HaluEvalExtractor",
    "hallucination_eval": f"{base_import}halueval:HaluEvalExtractor",

    # Agent benchmarks (NOT lm-eval)
    "agentbench": f"{base_import}agentbench:AgentBenchExtractor",
    "agent_bench": f"{base_import}agentbench:AgentBenchExtractor",
    "agentinstruct": f"{base_import}agentbench:AgentBenchExtractor",

    # Math Olympiad benchmarks (NOT lm-eval)
    "olympiadbench": f"{base_import}olympiadbench:OlympiadBenchExtractor",
    "olympiad_bench": f"{base_import}olympiadbench:OlympiadBenchExtractor",
    "imo_answerbench": f"{base_import}olympiadbench:OlympiadBenchExtractor",
    "imo_math": f"{base_import}olympiadbench:OlympiadBenchExtractor",

    # Chinese benchmarks (NOT lm-eval)
    "cluewsc": f"{base_import}cluewsc:CLUEWSCExtractor",
    "cluewsc2020": f"{base_import}cluewsc:CLUEWSCExtractor",
    "clue_wsc": f"{base_import}cluewsc:CLUEWSCExtractor",
    "chinese_simpleqa": f"{base_import}chinese_simpleqa:ChineseSimpleQAExtractor",
    "c_simpleqa": f"{base_import}chinese_simpleqa:ChineseSimpleQAExtractor",
    "csimpleqa": f"{base_import}chinese_simpleqa:ChineseSimpleQAExtractor",

    # Competitive programming benchmarks (NOT lm-eval)
    "codeforces": f"{base_import}codeforces:CodeforcesExtractor",
    "code_forces": f"{base_import}codeforces:CodeforcesExtractor",
    "codeelo": f"{base_import}codeforces:CodeforcesExtractor",

    # Code editing benchmarks (NOT lm-eval)
    "aider_polyglot": f"{base_import}aider_polyglot:AiderPolyglotExtractor",
    "aider-polyglot": f"{base_import}aider_polyglot:AiderPolyglotExtractor",
    "polyglot": f"{base_import}aider_polyglot:AiderPolyglotExtractor",
    "code_exercises": f"{base_import}aider_polyglot:AiderPolyglotExtractor",

    # Agent safety benchmarks (NOT lm-eval)
    "agentharm": f"{base_import}agentharm:AgentHarmExtractor",
    "agent_harm": f"{base_import}agentharm:AgentHarmExtractor",
    "agent-harm": f"{base_import}agentharm:AgentHarmExtractor",

    # Safety refusal benchmarks (NOT lm-eval)
    "donotanswer": f"{base_import}donotanswer:DoNotAnswerExtractor",
    "do_not_answer": f"{base_import}donotanswer:DoNotAnswerExtractor",
    "do-not-answer": f"{base_import}donotanswer:DoNotAnswerExtractor",
    "wildguard": f"{base_import}wildguard:WildGuardExtractor",
    "wildguardmix": f"{base_import}wildguard:WildGuardExtractor",
    "wild_guard": f"{base_import}wildguard:WildGuardExtractor",

    # Hallucination benchmarks (NOT lm-eval)
    "faithbench": f"{base_import}faithbench:FaithBenchExtractor",
    "faith_bench": f"{base_import}faithbench:FaithBenchExtractor",
    "faith-bench": f"{base_import}faithbench:FaithBenchExtractor",

    # Sycophancy benchmarks (NOT lm-eval)
    "sycophancy_eval": f"{base_import}sycophancy_eval:SycophancyEvalExtractor",
    "sycophancy-eval": f"{base_import}sycophancy_eval:SycophancyEvalExtractor",
    "syceval": f"{base_import}sycophancy_eval:SycophancyEvalExtractor",
    "syc_eval": f"{base_import}sycophancy_eval:SycophancyEvalExtractor",

    # Agent task benchmarks (NOT lm-eval)
    "tau_bench": f"{base_import}tau_bench:TauBenchExtractor",
    "tau-bench": f"{base_import}tau_bench:TauBenchExtractor",
    "taubench": f"{base_import}tau_bench:TauBenchExtractor",
    "tau2_bench": f"{base_import}tau_bench:TauBenchExtractor",

    # Agent safety/emulation benchmarks (NOT lm-eval)
    "toolemu": f"{base_import}toolemu:ToolEmuExtractor",
    "tool_emu": f"{base_import}toolemu:ToolEmuExtractor",
    "tool-emu": f"{base_import}toolemu:ToolEmuExtractor",

    # Function calling/API benchmarks (NOT lm-eval)
    "toolbench": f"{base_import}toolbench:ToolBenchExtractor",
    "tool_bench": f"{base_import}toolbench:ToolBenchExtractor",
    "toolllm": f"{base_import}toolbench:ToolBenchExtractor",
    "tool_llm": f"{base_import}toolbench:ToolBenchExtractor",
    "nexus_function_calling": f"{base_import}toolbench:ToolBenchExtractor",

    # Planning benchmarks (NOT lm-eval)
    "travelplanner": f"{base_import}travelplanner:TravelPlannerExtractor",
    "travel_planner": f"{base_import}travelplanner:TravelPlannerExtractor",
    "travel-planner": f"{base_import}travelplanner:TravelPlannerExtractor",

    # Factuality/Grounding benchmarks (NOT lm-eval)
    "facts_grounding": f"{base_import}facts_grounding:FACTSGroundingExtractor",
    "facts-grounding": f"{base_import}facts_grounding:FACTSGroundingExtractor",
    "factsgrounding": f"{base_import}facts_grounding:FACTSGroundingExtractor",

    # Web browsing benchmarks (NOT lm-eval)
    "browsecomp": f"{base_import}agentic_search:BrowseCompExtractor",
    "browse_comp": f"{base_import}agentic_search:BrowseCompExtractor",
    "browse-comp": f"{base_import}agentic_search:BrowseCompExtractor",
    "browsecomp_zh": f"{base_import}agentic_search:BrowseCompExtractor",

    # Selective refusal benchmarks (NOT lm-eval)
    "refusalbench": f"{base_import}refusalbench:RefusalBenchExtractor",
    "refusal_bench": f"{base_import}refusalbench:RefusalBenchExtractor",
    "refusal-bench": f"{base_import}refusalbench:RefusalBenchExtractor",

    # Chinese value alignment benchmarks (NOT lm-eval)
    "flames": f"{base_import}flames:FlamesExtractor",
    "flames_chinese": f"{base_import}flames:FlamesExtractor",

    # Planning/Reasoning benchmarks (NOT lm-eval)
    "planbench": f"{base_import}planbench:PlanBenchExtractor",
    "plan_bench": f"{base_import}planbench:PlanBenchExtractor",
    "planningbench": f"{base_import}planbench:PlanBenchExtractor",

    # SWE-bench variants (NOT lm-eval)
    "swe_verified": f"{base_import}swe_bench:SWEBenchVerifiedExtractor",
    "swe_bench_verified": f"{base_import}swe_bench:SWEBenchVerifiedExtractor",
    "swebench_verified": f"{base_import}swe_bench:SWEBenchVerifiedExtractor",
    "multi_swe_bench": f"{base_import}swe_bench:MultiSWEBenchExtractor",
    "multiswebench": f"{base_import}swe_bench:MultiSWEBenchExtractor",
    "swe_bench_multilingual": f"{base_import}swe_bench:MultiSWEBenchExtractor",

    # LiveCodeBench V6 (NOT lm-eval)
    "livecodebench_v6": f"{base_import}livecodebench_v6:LiveCodeBenchV6Extractor",
    "livecodebench_v5": f"{base_import}livecodebench_v6:LiveCodeBenchV6Extractor",
    "livecodebench_lite": f"{base_import}livecodebench_v6:LiveCodeBenchV6Extractor",

    # Hallucinations Leaderboard (NOT lm-eval)
    "hallucinations_leaderboard": f"{base_import}hallucinations_leaderboard:HallucinationsLeaderboardExtractor",
    "hallucination_leaderboard": f"{base_import}hallucinations_leaderboard:HallucinationsLeaderboardExtractor",
    "hallu_leaderboard": f"{base_import}hallucinations_leaderboard:HallucinationsLeaderboardExtractor",

    # Longform Writing (NOT lm-eval)
    "longform_writing": f"{base_import}longform_writing:LongformWritingExtractor",
    "longform": f"{base_import}longform_writing:LongformWritingExtractor",
    "longform_generation": f"{base_import}longform_writing:LongformWritingExtractor",

    # Agentic Search benchmarks (NOT lm-eval)
    "seal_0": f"{base_import}agentic_search:SealExtractor",
    "seal": f"{base_import}agentic_search:SealExtractor",
    "finsearchcomp": f"{base_import}agentic_search:FinSearchCompExtractor",
    "fin_search_comp": f"{base_import}agentic_search:FinSearchCompExtractor",
    "financial_search": f"{base_import}agentic_search:FinSearchCompExtractor",

    # Coding benchmarks (NOT lm-eval)
    "oj_bench": f"{base_import}coding_benchmarks:OJBenchExtractor",
    "ojbench": f"{base_import}coding_benchmarks:OJBenchExtractor",
    "online_judge_bench": f"{base_import}coding_benchmarks:OJBenchExtractor",
    "terminal_bench": f"{base_import}coding_benchmarks:TerminalBenchExtractor",
    "terminalbench": f"{base_import}coding_benchmarks:TerminalBenchExtractor",
    "terminal": f"{base_import}coding_benchmarks:TerminalBenchExtractor",
    # scicode and sci_code are already defined earlier - use scicode.py extractor

    # Medium priority benchmarks (NOT lm-eval)
    "cnmo_2024": f"{base_import}medium_priority_benchmarks:CNMOExtractor",
    "cnmo": f"{base_import}medium_priority_benchmarks:CNMOExtractor",
    "chinese_math_olympiad": f"{base_import}medium_priority_benchmarks:CNMOExtractor",
    "curate": f"{base_import}medium_priority_benchmarks:CurateExtractor",
    "personalized_alignment": f"{base_import}medium_priority_benchmarks:CurateExtractor",
    "halulens": f"{base_import}medium_priority_benchmarks:HalulensExtractor",
    "hallu_lens": f"{base_import}medium_priority_benchmarks:HalulensExtractor",
    "hallucination_lens": f"{base_import}medium_priority_benchmarks:HalulensExtractor",
    "politicalbias_qa": f"{base_import}medium_priority_benchmarks:PoliticalBiasExtractor",
    "political_bias": f"{base_import}medium_priority_benchmarks:PoliticalBiasExtractor",
    "politicalbias": f"{base_import}medium_priority_benchmarks:PoliticalBiasExtractor",
    "polyglottoxicityprompts": f"{base_import}medium_priority_benchmarks:PolygloToxicityExtractor",
    "polyglot_toxicity": f"{base_import}medium_priority_benchmarks:PolygloToxicityExtractor",
    "multilingual_toxicity": f"{base_import}medium_priority_benchmarks:PolygloToxicityExtractor",
    # Custom TruthfulQA (loads directly from HF)
    "truthfulqa_custom": f"{base_import}truthfulqa_custom:TruthfulQACustomExtractor",

    # Okapi Multilingual benchmarks
    "okapi_mmlu_multilingual": f"{base_import}okapi_multilingual:OkapiMMLUExtractor",
    "okapi_mmlu": f"{base_import}okapi_multilingual:OkapiMMLUExtractor",
    "okapi_hellaswag_multilingual": f"{base_import}okapi_multilingual:OkapiHellaswagExtractor",
    "okapi_hellaswag": f"{base_import}okapi_multilingual:OkapiHellaswagExtractor",
    "okapi_truthfulqa_multilingual": f"{base_import}okapi_multilingual:OkapiTruthfulQAExtractor",
    "okapi_truthfulqa": f"{base_import}okapi_multilingual:OkapiTruthfulQAExtractor",

    # Multilingual benchmarks
    "paws_x": f"{base_import}multilingual_benchmarks:PawsXExtractor",
    "paws-x": f"{base_import}multilingual_benchmarks:PawsXExtractor",
    "mlqa": f"{base_import}multilingual_benchmarks:MLQAExtractor",
    "darija_bench": f"{base_import}multilingual_benchmarks:DarijaBenchExtractor",
    "eus_exams": f"{base_import}multilingual_benchmarks:EusExamsExtractor",
    "lambada_multilingual_stablelm": f"{base_import}multilingual_benchmarks:LambadaMultilingualExtractor",
    "lambada_multilingual": f"{base_import}multilingual_benchmarks:LambadaMultilingualExtractor",

    # Reasoning benchmarks
    "inverse_scaling": f"{base_import}reasoning_benchmarks:InverseScalingExtractor",

    # Medical benchmarks
    "med_concepts_qa": f"{base_import}medical_benchmarks:MedConceptsQAExtractor",

    # Translation benchmarks
    "translation": f"{base_import}translation_benchmarks:TranslationExtractor",
    "wmt14_en_fr": f"{base_import}translation_benchmarks:WMT14Extractor",
    "wmt14_fr_en": f"{base_import}translation_benchmarks:WMT14Extractor",
    "wmt16_de_en": f"{base_import}translation_benchmarks:WMT16Extractor",
    "wmt16_en_de": f"{base_import}translation_benchmarks:WMT16Extractor",
    "wmt2016": f"{base_import}translation_benchmarks:WMT16Extractor",

    # Multimodal benchmarks
    "mmmu": f"{base_import}multimodal_benchmarks:MMMUExtractor",
}

# Alias for backwards compatibility
HF_EXTRACTORS = EXTRACTORS




"""
Nonsense baseline for validating that metrics detect real signal.

Generate random token sequences to create a null distribution,
then compare benchmark metrics against this baseline.
"""

import torch
import numpy as np
import random
from typing import Dict, Tuple, Any


def generate_nonsense_activations(
    model,
    tokenizer,
    n_pairs: int = 50,
    layer: int = None,
    device: str = "cuda",
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Generate activations from random token sequences (nonsense baseline).
    
    This creates a null baseline - what activation differences look like
    when there is NO semantic concept.
    """
    vocab_size = tokenizer.vocab_size
    if layer is None:
        layer = model.config.num_hidden_layers // 2
    
    def generate_random_tokens(min_tokens=5, max_tokens=25):
        n_tokens = random.randint(min_tokens, max_tokens)
        token_ids = [random.randint(100, vocab_size - 100) for _ in range(n_tokens)]
        return tokenizer.decode(token_ids)
    
    def get_activation(text):
        inputs = tokenizer(text, return_tensors="pt", truncation=True, max_length=256)
        inputs = {k: v.to(device) for k, v in inputs.items()}
        with torch.no_grad():
            outputs = model(**inputs, output_hidden_states=True)
            hs = outputs.hidden_states[layer + 1]
            return hs[0, -1, :]
    
    pos_acts = []
    neg_acts = []
    
    for _ in range(n_pairs):
        try:
            pos_text = generate_random_tokens()
            neg_text = generate_random_tokens()
            pos_acts.append(get_activation(pos_text))
            neg_acts.append(get_activation(neg_text))
        except Exception:
            continue
    
    if len(pos_acts) < 10:
        raise ValueError(f"Could only generate {len(pos_acts)} nonsense pairs")
    
    return torch.stack(pos_acts), torch.stack(neg_acts)


def compute_nonsense_baseline(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
    nonsense_pos: torch.Tensor,
    nonsense_neg: torch.Tensor,
    n_folds: int = 5,
) -> Dict[str, float]:
    """
    Compare metrics against nonsense baseline.
    
    Returns z-scores: how many standard deviations above nonsense.
    """
    from .probe_metrics import compute_linear_probe_accuracy, compute_signal_strength
    
    real_linear = compute_linear_probe_accuracy(pos_activations, neg_activations, n_folds)
    real_signal = compute_signal_strength(pos_activations, neg_activations, n_folds)
    
    nonsense_linear = compute_linear_probe_accuracy(nonsense_pos, nonsense_neg, n_folds)
    nonsense_signal = compute_signal_strength(nonsense_pos, nonsense_neg, n_folds)
    
    linear_z = (real_linear - nonsense_linear) / 0.1  # Assume std ~0.1
    signal_z = (real_signal - nonsense_signal) / 0.1
    
    return {
        "real_linear_probe": real_linear,
        "real_signal_strength": real_signal,
        "nonsense_linear_probe": nonsense_linear,
        "nonsense_signal_strength": nonsense_signal,
        "linear_z_score": float(linear_z),
        "signal_z_score": float(signal_z),
        "is_significant": linear_z > 2.0 or signal_z > 2.0,
    }


def analyze_with_nonsense_baseline(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
    model=None,
    tokenizer=None,
    layer: int = None,
    device: str = "cuda",
) -> Dict[str, Any]:
    """
    Full analysis comparing real data to nonsense baseline.
    """
    from .probe_metrics import (
        compute_linear_probe_accuracy,
        compute_signal_strength,
        compute_mlp_probe_accuracy,
    )
    from .icd import compute_icd
    
    real_metrics = {
        "linear_probe": compute_linear_probe_accuracy(pos_activations, neg_activations),
        "signal_strength": compute_signal_strength(pos_activations, neg_activations),
        "mlp_probe": compute_mlp_probe_accuracy(pos_activations, neg_activations),
        "icd": compute_icd(pos_activations, neg_activations)["icd"],
    }
    
    if model is not None and tokenizer is not None:
        try:
            nonsense_pos, nonsense_neg = generate_nonsense_activations(
                model, tokenizer, n_pairs=50, layer=layer, device=device
            )
            
            nonsense_metrics = {
                "linear_probe": compute_linear_probe_accuracy(nonsense_pos, nonsense_neg),
                "signal_strength": compute_signal_strength(nonsense_pos, nonsense_neg),
                "mlp_probe": compute_mlp_probe_accuracy(nonsense_pos, nonsense_neg),
                "icd": compute_icd(nonsense_pos, nonsense_neg)["icd"],
            }
            
            z_scores = {}
            for key in real_metrics:
                diff = real_metrics[key] - nonsense_metrics[key]
                z_scores[f"{key}_z"] = diff / 0.1
            
            return {
                "real": real_metrics,
                "nonsense": nonsense_metrics,
                "z_scores": z_scores,
                "is_real_signal": z_scores.get("linear_probe_z", 0) > 2.0,
            }
        except Exception as e:
            return {
                "real": real_metrics,
                "nonsense": None,
                "error": str(e),
            }
    
    return {"real": real_metrics, "nonsense": None}



from __future__ import annotations

from typing import Any
from wisent.core.cli_logger import setup_logger

from wisent.core.contrastive_pairs.core.pair import ContrastivePair
from wisent.core.contrastive_pairs.core.response import NegativeResponse, PositiveResponse
from wisent.core.contrastive_pairs.huggingface_pairs.atoms import HuggingFaceBenchmarkExtractor

from sympy.parsing.latex import parse_latex
from sympy import latex

__all__ = ["PolyMathExtractor"]

log = setup_logger(__name__)

task_names = (
    "polymath_ar_top", "polymath_ar_high", "polymath_ar_medium", "polymath_ar_low",
    "polymath_bn_top", "polymath_bn_high", "polymath_bn_medium", "polymath_bn_low",
    "polymath_de_top", "polymath_de_high", "polymath_de_medium", "polymath_de_low",
    "polymath_en_top", "polymath_en_high", "polymath_en_medium", "polymath_en_low",
    "polymath_es_top", "polymath_es_high", "polymath_es_medium", "polymath_es_low",
    "polymath_fr_top", "polymath_fr_high", "polymath_fr_medium", "polymath_fr_low",
    "polymath_id_top", "polymath_id_high", "polymath_id_medium", "polymath_id_low",
    "polymath_it_top", "polymath_it_high", "polymath_it_medium", "polymath_it_low",
    "polymath_ja_top", "polymath_ja_high", "polymath_ja_medium", "polymath_ja_low",
    "polymath_ko_top", "polymath_ko_high", "polymath_ko_medium", "polymath_ko_low",
    "polymath_ms_top", "polymath_ms_high", "polymath_ms_medium", "polymath_ms_low",
    "polymath_pt_top", "polymath_pt_high", "polymath_pt_medium", "polymath_pt_low",
    "polymath_ru_top", "polymath_ru_high", "polymath_ru_medium", "polymath_ru_low",
    "polymath_sw_top", "polymath_sw_high", "polymath_sw_medium", "polymath_sw_low",
    "polymath_te_top", "polymath_te_high", "polymath_te_medium", "polymath_te_low",
    "polymath_th_top", "polymath_th_high", "polymath_th_medium", "polymath_th_low",
    "polymath_vi_top", "polymath_vi_high", "polymath_vi_medium", "polymath_vi_low",
    "polymath_zh_top", "polymath_zh_high", "polymath_zh_medium", "polymath_zh_low",
)

class PolyMathExtractor(HuggingFaceBenchmarkExtractor):
    """
    Extractor for PolyMath dataset (multilingual mathematical reasoning).

    PolyMath schema (Qwen/PolyMath):
        - question: str (math problem statement in Chinese or English)
        - answer: str (final answer)
    """


    evaluator_name = "polymath"

    # These will be overridden by subclasses
    language = "en"
    difficulty = "medium"

    def extract_contrastive_pairs(
        self,
        limit: int | None = None,
    ) -> list[ContrastivePair]:
        """
        Build contrastive pairs from PolyMath examples.

        Args:
            limit: Optional maximum number of pairs to produce.

        Returns:
            A list of ContrastivePair objects.
        """
        max_items = self._normalize_limit(limit)

        # Load PolyMath dataset with appropriate config and split
        # Config is the language (e.g., "en", "zh")
        # Split is the difficulty level (e.g., "high", "medium", "low", "top")
        docs = self.load_dataset(
            dataset_name="Qwen/PolyMath",
            dataset_config=self.language,
            split=self.difficulty,
            limit=max_items,
        )

        pairs: list[ContrastivePair] = []

        log.info(f"Extracting contrastive pairs from {len(docs)} PolyMath ({self.language}_{self.difficulty}) examples")

        for doc in docs:
            pair = self._extract_pair_from_doc(doc)
            if pair is not None:
                pairs.append(pair)
                if max_items is not None and len(pairs) >= max_items:
                    break

        if not pairs:
            log.warning(f"No valid PolyMath ({self.language}_{self.difficulty}) pairs extracted")

        return pairs

    def _extract_pair_from_doc(self, doc: dict[str, Any]) -> ContrastivePair | None:
        """
        Convert a single PolyMath doc into a ContrastivePair.

        Returns None when required fields are missing or malformed.
        """
        try:
            problem = doc.get("question", "").strip()
            correct = doc.get("answer", "").strip()

            if not problem or not correct:
                log.debug("Skipping: missing problem or answer")
                return None

            incorrect_answer = self._create_incorrect_answer(correct)

            # Format the question
            question = f"Question: {problem}\n\nWhat is the answer?"

            metadata = {
                "label": "polymath",
                "source": "Qwen/PolyMath",
                "language": self.language,
                "difficulty": self.difficulty,
            }

            return self._build_pair(
                question=question,
                correct=correct,
                incorrect=incorrect_answer,
                metadata=metadata,
            )

        except Exception as exc:
            log.error(f"Error extracting pair from doc: {exc}", exc_info=True)
            return None

    def _create_incorrect_answer(self, correct: str) -> str:
        """Create a meaningful incorrect answer using plausible wrong values."""
        import random
        import re
        random.seed(hash(correct) % (2**32))

        # Try symbolic parsing first
        try:
            parsed_correct = parse_latex(correct)
            transforms = [
                parsed_correct * 2,
                parsed_correct / 2,
                parsed_correct - 1,
                -parsed_correct,
            ]
            wrong = random.choice(transforms)
            return str(latex(wrong))
        except Exception:
            pass

        # Try simple integer
        try:
            clean = correct.replace('$', '').replace(',', '').strip()
            num = int(clean)
            wrong_vals = [num * 2, num // 2 if num > 1 else num * 3, num - 1, -num]
            return str(random.choice(wrong_vals))
        except ValueError:
            pass

        # For fractions
        frac_match = re.match(r'\\frac\{(\d+)\}\{(\d+)\}', correct)
        if frac_match:
            n, d = int(frac_match.group(1)), int(frac_match.group(2))
            return random.choice([f"\\frac{{{d}}}{{{n}}}", f"\\frac{{{n*2}}}{{{d}}}"])

        # For pi expressions
        if '\\pi' in correct:
            return correct.replace('\\pi', '2\\pi') if '2\\pi' not in correct else correct.replace('2\\pi', '\\pi')

        # Fallback
        return random.choice(['0', '1', '-1', '2'])


    @staticmethod
    def _build_pair(
        question: str,
        correct: str,
        incorrect: str,
        metadata: dict[str, Any] | None = None,
    ) -> ContrastivePair:
        """Build a ContrastivePair from question and responses."""
        positive_response = PositiveResponse(model_response=correct)
        negative_response = NegativeResponse(model_response=incorrect)
        return ContrastivePair(
            prompt=question,
            positive_response=positive_response,
            negative_response=negative_response,
            label=metadata.get("label") if metadata else None,
            metadata=metadata,
        )


# All 72 language/difficulty specific extractors

# Arabic
class PolyMathARTopExtractor(PolyMathExtractor):
    language = "ar"
    difficulty = "top"

class PolyMathARHighExtractor(PolyMathExtractor):
    language = "ar"
    difficulty = "high"

class PolyMathARMediumExtractor(PolyMathExtractor):
    language = "ar"
    difficulty = "medium"

class PolyMathARLowExtractor(PolyMathExtractor):
    language = "ar"
    difficulty = "low"

# Bengali
class PolyMathBNTopExtractor(PolyMathExtractor):
    language = "bn"
    difficulty = "top"

class PolyMathBNHighExtractor(PolyMathExtractor):
    language = "bn"
    difficulty = "high"

class PolyMathBNMediumExtractor(PolyMathExtractor):
    language = "bn"
    difficulty = "medium"

class PolyMathBNLowExtractor(PolyMathExtractor):
    language = "bn"
    difficulty = "low"

# German
class PolyMathDETopExtractor(PolyMathExtractor):
    language = "de"
    difficulty = "top"

class PolyMathDEHighExtractor(PolyMathExtractor):
    language = "de"
    difficulty = "high"

class PolyMathDEMediumExtractor(PolyMathExtractor):
    language = "de"
    difficulty = "medium"

class PolyMathDELowExtractor(PolyMathExtractor):
    language = "de"
    difficulty = "low"

# English
class PolyMathENTopExtractor(PolyMathExtractor):
    language = "en"
    difficulty = "top"

class PolyMathENHighExtractor(PolyMathExtractor):
    language = "en"
    difficulty = "high"

class PolyMathENMediumExtractor(PolyMathExtractor):
    language = "en"
    difficulty = "medium"

class PolyMathENLowExtractor(PolyMathExtractor):
    language = "en"
    difficulty = "low"

# Spanish
class PolyMathESTopExtractor(PolyMathExtractor):
    language = "es"
    difficulty = "top"

class PolyMathESHighExtractor(PolyMathExtractor):
    language = "es"
    difficulty = "high"

class PolyMathESMediumExtractor(PolyMathExtractor):
    language = "es"
    difficulty = "medium"

class PolyMathESLowExtractor(PolyMathExtractor):
    language = "es"
    difficulty = "low"

# French
class PolyMathFRTopExtractor(PolyMathExtractor):
    language = "fr"
    difficulty = "top"

class PolyMathFRHighExtractor(PolyMathExtractor):
    language = "fr"
    difficulty = "high"

class PolyMathFRMediumExtractor(PolyMathExtractor):
    language = "fr"
    difficulty = "medium"

class PolyMathFRLowExtractor(PolyMathExtractor):
    language = "fr"
    difficulty = "low"

# Indonesian
class PolyMathIDTopExtractor(PolyMathExtractor):
    language = "id"
    difficulty = "top"

class PolyMathIDHighExtractor(PolyMathExtractor):
    language = "id"
    difficulty = "high"

class PolyMathIDMediumExtractor(PolyMathExtractor):
    language = "id"
    difficulty = "medium"

class PolyMathIDLowExtractor(PolyMathExtractor):
    language = "id"
    difficulty = "low"

# Italian
class PolyMathITTopExtractor(PolyMathExtractor):
    language = "it"
    difficulty = "top"

class PolyMathITHighExtractor(PolyMathExtractor):
    language = "it"
    difficulty = "high"

class PolyMathITMediumExtractor(PolyMathExtractor):
    language = "it"
    difficulty = "medium"

class PolyMathITLowExtractor(PolyMathExtractor):
    language = "it"
    difficulty = "low"

# Japanese
class PolyMathJATopExtractor(PolyMathExtractor):
    language = "ja"
    difficulty = "top"

class PolyMathJAHighExtractor(PolyMathExtractor):
    language = "ja"
    difficulty = "high"

class PolyMathJAMediumExtractor(PolyMathExtractor):
    language = "ja"
    difficulty = "medium"

class PolyMathJALowExtractor(PolyMathExtractor):
    language = "ja"
    difficulty = "low"

# Korean
class PolyMathKOTopExtractor(PolyMathExtractor):
    language = "ko"
    difficulty = "top"

class PolyMathKOHighExtractor(PolyMathExtractor):
    language = "ko"
    difficulty = "high"

class PolyMathKOMediumExtractor(PolyMathExtractor):
    language = "ko"
    difficulty = "medium"

class PolyMathKOLowExtractor(PolyMathExtractor):
    language = "ko"
    difficulty = "low"

# Malay
class PolyMathMSTopExtractor(PolyMathExtractor):
    language = "ms"
    difficulty = "top"

class PolyMathMSHighExtractor(PolyMathExtractor):
    language = "ms"
    difficulty = "high"

class PolyMathMSMediumExtractor(PolyMathExtractor):
    language = "ms"
    difficulty = "medium"

class PolyMathMSLowExtractor(PolyMathExtractor):
    language = "ms"
    difficulty = "low"

# Portuguese
class PolyMathPTTopExtractor(PolyMathExtractor):
    language = "pt"
    difficulty = "top"

class PolyMathPTHighExtractor(PolyMathExtractor):
    language = "pt"
    difficulty = "high"

class PolyMathPTMediumExtractor(PolyMathExtractor):
    language = "pt"
    difficulty = "medium"

class PolyMathPTLowExtractor(PolyMathExtractor):
    language = "pt"
    difficulty = "low"

# Russian
class PolyMathRUTopExtractor(PolyMathExtractor):
    language = "ru"
    difficulty = "top"

class PolyMathRUHighExtractor(PolyMathExtractor):
    language = "ru"
    difficulty = "high"

class PolyMathRUMediumExtractor(PolyMathExtractor):
    language = "ru"
    difficulty = "medium"

class PolyMathRULowExtractor(PolyMathExtractor):
    language = "ru"
    difficulty = "low"

# Swahili
class PolyMathSWTopExtractor(PolyMathExtractor):
    language = "sw"
    difficulty = "top"

class PolyMathSWHighExtractor(PolyMathExtractor):
    language = "sw"
    difficulty = "high"

class PolyMathSWMediumExtractor(PolyMathExtractor):
    language = "sw"
    difficulty = "medium"

class PolyMathSWLowExtractor(PolyMathExtractor):
    language = "sw"
    difficulty = "low"

# Telugu
class PolyMathTETopExtractor(PolyMathExtractor):
    language = "te"
    difficulty = "top"

class PolyMathTEHighExtractor(PolyMathExtractor):
    language = "te"
    difficulty = "high"

class PolyMathTEMediumExtractor(PolyMathExtractor):
    language = "te"
    difficulty = "medium"

class PolyMathTELowExtractor(PolyMathExtractor):
    language = "te"
    difficulty = "low"

# Thai
class PolyMathTHTopExtractor(PolyMathExtractor):
    language = "th"
    difficulty = "top"

class PolyMathTHHighExtractor(PolyMathExtractor):
    language = "th"
    difficulty = "high"

class PolyMathTHMediumExtractor(PolyMathExtractor):
    language = "th"
    difficulty = "medium"

class PolyMathTHLowExtractor(PolyMathExtractor):
    language = "th"
    difficulty = "low"

# Vietnamese
class PolyMathVITopExtractor(PolyMathExtractor):
    language = "vi"
    difficulty = "top"

class PolyMathVIHighExtractor(PolyMathExtractor):
    language = "vi"
    difficulty = "high"

class PolyMathVIMediumExtractor(PolyMathExtractor):
    language = "vi"
    difficulty = "medium"

class PolyMathVILowExtractor(PolyMathExtractor):
    language = "vi"
    difficulty = "low"

# Chinese
class PolyMathZHTopExtractor(PolyMathExtractor):
    language = "zh"
    difficulty = "top"

class PolyMathZHHighExtractor(PolyMathExtractor):
    language = "zh"
    difficulty = "high"

class PolyMathZHMediumExtractor(PolyMathExtractor):
    language = "zh"
    difficulty = "medium"

class PolyMathZHLowExtractor(PolyMathExtractor):
    language = "zh"
    difficulty = "low"



"""Database loaders for activations and pair texts from Supabase with local caching."""
from typing import Dict, Optional, List, Tuple
import os
import struct
import json
import torch
from .cache import get_cache_path, save_pair_texts_cache, save_activations_cache


def _get_db_connection(database_url: Optional[str] = None):
    """Get database connection with proper URL handling."""
    try:
        import psycopg2
    except ImportError:
        raise ImportError("psycopg2 required for database loading. Install with: pip install psycopg2-binary")

    db_url = database_url or os.environ.get("DATABASE_URL")
    if not db_url:
        raise ValueError("No database URL provided. Set DATABASE_URL env var or pass database_url parameter.")

    # Add sslmode if not present
    if "sslmode=" not in db_url:
        db_url += "?sslmode=require" if "?" not in db_url else "&sslmode=require"

    print(f"  DEBUG: Connecting to database...", flush=True)
    conn = psycopg2.connect(db_url)
    cur = conn.cursor()
    cur.execute("SET statement_timeout = '600s'")  # 10 minutes for large queries
    cur.close()
    print(f"  DEBUG: Connected successfully", flush=True)
    return conn


def _bytes_to_vector_simple(data: bytes) -> List[float]:
    """Convert bytes to activation vector (pre-extracted, single vector)."""
    num_floats = len(data) // 4
    return list(struct.unpack(f'{num_floats}f', data))


def load_activations_from_database(
    model_name: str,
    task_name: str,
    layer: int,
    prompt_format: str = "chat",
    extraction_strategy: str = "last_token",
    limit: Optional[int] = None,
    database_url: Optional[str] = None,
    pair_ids: Optional[set] = None,
    use_cache: bool = True,
    force_refresh: bool = False,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Load activations for a single layer from local cache or database.

    Args:
        model_name: HuggingFace model ID (e.g., "meta-llama/Llama-3.2-1B-Instruct")
        task_name: Name of the contrastive pair set (e.g., "truthfulqa_custom")
        layer: Layer number to load
        prompt_format: Prompt format used ("chat" or "completion")
        extraction_strategy: Token extraction strategy ("last_token" or "first_token")
        limit: Maximum number of pairs to load
        database_url: Optional database URL (defaults to DATABASE_URL env var)
        pair_ids: Optional set of pair IDs to filter (for train/test split)
        use_cache: Whether to use local cache (default: True)
        force_refresh: Force reload from database even if cache exists

    Returns:
        Tuple of (pos_activations, neg_activations) tensors with pair_id mapping
    """
    cache_path = get_cache_path(task_name, "activations", model_name=model_name, layer=layer)
    print(f"  DEBUG: activations cache_path = {cache_path}", flush=True)
    print(f"  DEBUG: cache_path.exists() = {cache_path.exists()}", flush=True)

    # Try loading from cache first
    if use_cache and not force_refresh and cache_path.exists():
        print(f"  Loading activations from cache: {cache_path}", flush=True)
        cached = torch.load(cache_path, weights_only=True)
        pos_tensor = cached["pos"]
        neg_tensor = cached["neg"]
        cached_pair_ids = cached.get("pair_ids", list(range(len(pos_tensor))))
        print(f"  DEBUG: loaded {len(pos_tensor)} activation pairs from cache", flush=True)

        # Filter by pair_ids if specified
        if pair_ids is not None:
            pos_list, neg_list = [], []
            for i, pid in enumerate(cached_pair_ids):
                if pid in pair_ids:
                    pos_list.append(pos_tensor[i])
                    neg_list.append(neg_tensor[i])
            if pos_list:
                print(f"  DEBUG: filtered to {len(pos_list)} pairs", flush=True)
                return torch.stack(pos_list), torch.stack(neg_list)
            print(f"  DEBUG: no matching pairs in cache, falling through to database", flush=True)
        else:
            # Apply limit if specified
            if limit and len(pos_tensor) > limit:
                pos_tensor = pos_tensor[:limit]
                neg_tensor = neg_tensor[:limit]
            return pos_tensor, neg_tensor
    from collections import defaultdict

    print(f"  DEBUG: Loading activations from database (cache miss)", flush=True)
    conn = _get_db_connection(database_url)
    cur = conn.cursor()

    try:
        # Get model ID
        print(f"  DEBUG: Querying model ID...", flush=True)
        cur.execute('SELECT id FROM "Model" WHERE "huggingFaceId" = %s', (model_name,))
        result = cur.fetchone()
        if not result:
            raise ValueError(f"Model {model_name} not found in database")
        model_id = result[0]
        print(f"  DEBUG: Model ID = {model_id}", flush=True)

        # Get set ID
        print(f"  DEBUG: Querying set ID...", flush=True)
        cur.execute('SELECT id FROM "ContrastivePairSet" WHERE name = %s', (task_name,))
        result = cur.fetchone()
        if not result:
            raise ValueError(f"Benchmark {task_name} not found in database")
        set_id = result[0]
        print(f"  DEBUG: Set ID = {set_id}", flush=True)

        # Map CLI extraction strategy to database strategy name
        strategy_map = {"last_token": "chat_last", "first_token": "chat_first"}
        db_strategy = strategy_map.get(extraction_strategy, extraction_strategy)

        # Load pre-extracted activations from Activation table
        print(f"  DEBUG: Querying activations (strategy={db_strategy})...", flush=True)
        cur.execute('''
            SELECT "contrastivePairId", "activationData", "isPositive"
            FROM "Activation"
            WHERE "modelId" = %s
              AND "contrastivePairSetId" = %s
              AND "extractionStrategy" = %s
              AND layer = %s
            ORDER BY "contrastivePairId", "isPositive"
        ''', (model_id, set_id, db_strategy, layer))

        rows = cur.fetchall()
        print(f"  DEBUG: Fetched {len(rows)} rows", flush=True)

        # Group by pair_id
        pair_activations = defaultdict(dict)
        for row in rows:
            pair_id, activation_bytes, is_positive = row
            activation = _bytes_to_vector_simple(activation_bytes)
            key = "pos" if is_positive else "neg"
            pair_activations[pair_id][key] = activation

        # Convert to tensors - only pairs with both pos and neg
        pos_list, neg_list, loaded_pair_ids = [], [], []
        for pair_id, acts in sorted(pair_activations.items()):
            if "pos" in acts and "neg" in acts:
                pos_list.append(acts["pos"])
                neg_list.append(acts["neg"])
                loaded_pair_ids.append(pair_id)

        print(f"  DEBUG: Created tensors for {len(pos_list)} complete pairs", flush=True)
        all_pos_tensor = torch.tensor(pos_list, dtype=torch.float32)
        all_neg_tensor = torch.tensor(neg_list, dtype=torch.float32)

        # Save full data to cache
        if use_cache and len(pos_list) > 0:
            print(f"  Caching {len(pos_list)} activations to: {cache_path}", flush=True)
            torch.save({
                "pos": all_pos_tensor,
                "neg": all_neg_tensor,
                "pair_ids": loaded_pair_ids,
            }, cache_path)

        # Now filter by pair_ids if specified
        if pair_ids is not None:
            pos_filtered, neg_filtered = [], []
            for i, pid in enumerate(loaded_pair_ids):
                if pid in pair_ids:
                    pos_filtered.append(all_pos_tensor[i])
                    neg_filtered.append(all_neg_tensor[i])
            if pos_filtered:
                print(f"  DEBUG: Filtered to {len(pos_filtered)} pairs", flush=True)
                return torch.stack(pos_filtered), torch.stack(neg_filtered)
            return torch.tensor([]), torch.tensor([])

        # Apply limit if specified
        if limit and len(all_pos_tensor) > limit:
            all_pos_tensor = all_pos_tensor[:limit]
            all_neg_tensor = all_neg_tensor[:limit]

        return all_pos_tensor, all_neg_tensor

    finally:
        cur.close()
        conn.close()


def load_pair_texts_from_database(
    task_name: str,
    limit: int = 200,
    database_url: Optional[str] = None,
    use_cache: bool = True,
    force_refresh: bool = False,
) -> Dict[int, Dict[str, str]]:
    """
    Load contrastive pair texts from local cache or Supabase database.

    Args:
        task_name: Name of the contrastive pair set (e.g., "truthfulqa_custom")
        limit: Maximum number of pairs to load
        database_url: Optional database URL (defaults to DATABASE_URL env var)
        use_cache: Whether to use local cache (default: True)
        force_refresh: Force reload from database even if cache exists

    Returns:
        Dict mapping pair_id -> {prompt, positive, negative}
    """
    cache_path = get_cache_path(task_name, "pair_texts")

    # Try loading from cache first
    if use_cache and not force_refresh and cache_path.exists():
        print(f"  Loading pair texts from cache: {cache_path}", flush=True)
        with open(cache_path, 'r') as f:
            cached_data = json.load(f)
        pairs = {int(k): v for k, v in cached_data.items()}
        if limit and len(pairs) > limit:
            pair_ids = sorted(pairs.keys())[:limit]
            pairs = {pid: pairs[pid] for pid in pair_ids}
        return pairs

    # Load from database
    conn = _get_db_connection(database_url)
    cur = conn.cursor()

    try:
        # Get set ID for task
        cur.execute('SELECT id FROM "ContrastivePairSet" WHERE name = %s', (task_name,))
        result = cur.fetchone()
        if not result:
            raise ValueError(f"No ContrastivePairSet found with name '{task_name}'")
        set_id = result[0]

        # Load pairs (load all for caching, apply limit after)
        cur.execute('''
            SELECT id, "positiveExample", "negativeExample"
            FROM "ContrastivePair"
            WHERE "setId" = %s
            ORDER BY id
        ''', (set_id,))

        pairs = {}
        for row in cur.fetchall():
            pair_id, pos_example, neg_example = row

            # Parse prompt and responses
            if "\n\n" in pos_example:
                prompt = pos_example.rsplit("\n\n", 1)[0]
                pos_response = pos_example.rsplit("\n\n", 1)[1]
            else:
                prompt = pos_example
                pos_response = ""

            if "\n\n" in neg_example:
                neg_response = neg_example.rsplit("\n\n", 1)[1]
            else:
                neg_response = neg_example

            pairs[pair_id] = {
                "prompt": prompt,
                "positive": pos_response,
                "negative": neg_response,
            }

        # Save to cache
        if use_cache:
            print(f"  Caching {len(pairs)} pair texts to: {cache_path}")
            with open(cache_path, 'w') as f:
                json.dump(pairs, f)

        # Apply limit
        if limit and len(pairs) > limit:
            pair_ids = sorted(pairs.keys())[:limit]
            pairs = {pid: pairs[pid] for pid in pair_ids}

        return pairs

    finally:
        cur.close()
        conn.close()



"""Parser for the 'steering-viz' command - steering effect visualization."""


def setup_steering_viz_parser(parser):
    """Set up the steering-viz command parser."""
    parser.add_argument(
        "--model",
        type=str,
        default="meta-llama/Llama-3.2-1B-Instruct",
        help="Model name (default: meta-llama/Llama-3.2-1B-Instruct)"
    )
    parser.add_argument(
        "--task",
        type=str,
        required=True,
        help="Task/benchmark name in database (e.g., truthfulqa_custom)"
    )
    parser.add_argument(
        "--layer",
        type=int,
        default=12,
        help="Layer to visualize (default: 12)"
    )
    parser.add_argument(
        "--strength",
        type=float,
        default=1.0,
        help="Steering strength multiplier (default: 1.0)"
    )
    parser.add_argument(
        "--n-test-prompts",
        type=int,
        default=50,
        help="Number of test prompts to run (default: 50)"
    )
    parser.add_argument(
        "--limit",
        type=int,
        default=200,
        help="Maximum reference pairs to load (default: 200)"
    )
    parser.add_argument(
        "--prompt-format",
        type=str,
        default="chat",
        choices=["chat", "completion"],
        help="Prompt format (default: chat)"
    )
    parser.add_argument(
        "--extraction-strategy",
        type=str,
        default="last_token",
        choices=["last_token", "first_token"],
        help="Token extraction strategy (default: last_token)"
    )
    parser.add_argument(
        "--database-url",
        type=str,
        default=None,
        help="Database URL (default: DATABASE_URL env var)"
    )
    parser.add_argument(
        "--output",
        type=str,
        default="./steering_effect.png",
        help="Output PNG file path (default: ./steering_effect.png)"
    )
    parser.add_argument(
        "--compare-responses",
        action="store_true",
        help="Also generate and compare text responses (base vs steered)"
    )
    parser.add_argument(
        "--n-response-samples",
        type=int,
        default=5,
        help="Number of samples to show response comparison for (default: 5)"
    )
    parser.add_argument(
        "--max-new-tokens",
        type=int,
        default=100,
        help="Max tokens to generate per response (default: 100)"
    )



"""
Activation space structure metrics.

These metrics characterize the geometry of the activation space itself,
not just the difference vectors. They describe where pos and neg
point clouds live and their relationship.
"""

import torch
import numpy as np
from typing import Dict, Any, List, Tuple, Optional
from scipy import stats
from scipy.spatial.distance import pdist, squareform


def compute_cloud_shape(
    activations: torch.Tensor,
) -> Dict[str, Any]:
    """
    Characterize the shape of a single activation point cloud.

    Measures:
    - Dimension (intrinsic)
    - Spread (how dispersed)
    - Centering (distance from origin)
    - Sphericity (how sphere-like)
    """
    X = activations.float().cpu().numpy()
    n, d = X.shape

    if n < 3:
        return {"error": "need at least 3 points"}

    # Norms and centering
    norms = np.linalg.norm(X, axis=1)
    centroid = X.mean(axis=0)
    centroid_norm = np.linalg.norm(centroid)

    # Center the data
    X_centered = X - centroid

    # Covariance and eigenvalues
    cov = np.cov(X_centered.T)
    eigenvalues = np.linalg.eigvalsh(cov)
    eigenvalues = np.sort(eigenvalues)[::-1]
    eigenvalues = np.maximum(eigenvalues, 0)

    total_var = eigenvalues.sum()

    # Intrinsic dimension estimates
    if total_var > 0:
        cumsum = np.cumsum(eigenvalues) / total_var
        dim_50 = int(np.searchsorted(cumsum, 0.5) + 1)
        dim_90 = int(np.searchsorted(cumsum, 0.9) + 1)
        dim_99 = int(np.searchsorted(cumsum, 0.99) + 1)

        # Participation ratio
        participation_ratio = (total_var ** 2) / (np.sum(eigenvalues ** 2) + 1e-8)
    else:
        dim_50, dim_90, dim_99 = 1, 1, 1
        participation_ratio = 1

    # Sphericity: ratio of smallest to largest eigenvalue
    # High sphericity = roughly equal spread in all directions
    if eigenvalues[0] > 0:
        sphericity = eigenvalues[-1] / eigenvalues[0]
    else:
        sphericity = 0

    # Norm distribution
    norm_mean = float(norms.mean())
    norm_std = float(norms.std())
    norm_cv = norm_std / (norm_mean + 1e-8)  # coefficient of variation

    return {
        # Centering
        "centroid_norm": float(centroid_norm),
        "mean_norm": norm_mean,
        "norm_std": norm_std,
        "norm_cv": norm_cv,  # low = similar norms (sphere-like)

        # Intrinsic dimension
        "dims_for_50pct": dim_50,
        "dims_for_90pct": dim_90,
        "dims_for_99pct": dim_99,
        "participation_ratio": float(participation_ratio),

        # Shape
        "sphericity": float(sphericity),
        "top_eigenvalue_ratio": float(eigenvalues[0] / total_var) if total_var > 0 else 0,

        # Raw eigenvalues (top 10)
        "top_eigenvalues": eigenvalues[:10].tolist(),
    }


def compute_cone_fit(
    activations: torch.Tensor,
) -> Dict[str, Any]:
    """
    Test if activations lie on a cone from the origin.

    A cone means: all vectors point in similar directions
    but with varying magnitudes.
    """
    X = activations.float().cpu().numpy()
    n, d = X.shape

    if n < 3:
        return {"error": "need at least 3 points"}

    # Normalize to unit sphere
    norms = np.linalg.norm(X, axis=1, keepdims=True)
    valid = norms.squeeze() > 1e-8
    X_normalized = X[valid] / norms[valid]

    if len(X_normalized) < 3:
        return {"error": "too few valid points"}

    # Cone axis = mean direction
    mean_dir = X_normalized.mean(axis=0)
    mean_dir_norm = np.linalg.norm(mean_dir)

    if mean_dir_norm < 1e-8:
        return {"is_cone": False, "cone_concentration": 0}

    cone_axis = mean_dir / mean_dir_norm

    # Angular distance from cone axis
    cos_angles = X_normalized @ cone_axis
    angles = np.arccos(np.clip(cos_angles, -1, 1))
    angles_deg = np.degrees(angles)

    # Cone half-angle (contains 90% of points)
    cone_half_angle = float(np.percentile(angles_deg, 90))

    # Concentration: how tightly clustered around axis
    # 1 = all on axis, 0 = spread over sphere
    concentration = float(cos_angles.mean())

    return {
        "cone_axis_strength": float(mean_dir_norm),  # how strong the mean direction is
        "cone_concentration": concentration,
        "cone_half_angle_90pct": cone_half_angle,
        "angle_mean": float(angles_deg.mean()),
        "angle_std": float(angles_deg.std()),
        "angle_max": float(angles_deg.max()),
    }


def compute_sphere_fit(
    activations: torch.Tensor,
) -> Dict[str, Any]:
    """
    Test if activations lie on a sphere (constant distance from a center).
    """
    X = activations.float().cpu().numpy()
    n, d = X.shape

    if n < 3:
        return {"error": "need at least 3 points"}

    # Fit sphere: find center that minimizes variance of distances
    # Simple approach: use centroid as center
    centroid = X.mean(axis=0)

    # Distances from centroid
    distances = np.linalg.norm(X - centroid, axis=1)
    mean_radius = float(distances.mean())
    radius_std = float(distances.std())

    # Sphere fit quality: low std/mean = good sphere
    radius_cv = radius_std / (mean_radius + 1e-8)

    # Also check distances from origin
    origin_distances = np.linalg.norm(X, axis=1)
    origin_radius_mean = float(origin_distances.mean())
    origin_radius_std = float(origin_distances.std())
    origin_radius_cv = origin_radius_std / (origin_radius_mean + 1e-8)

    return {
        # Sphere around centroid
        "centroid_radius_mean": mean_radius,
        "centroid_radius_std": radius_std,
        "centroid_radius_cv": radius_cv,  # low = good sphere fit

        # Sphere around origin
        "origin_radius_mean": origin_radius_mean,
        "origin_radius_std": origin_radius_std,
        "origin_radius_cv": origin_radius_cv,
    }


def compute_manifold_dimension(
    activations: torch.Tensor,
    k_neighbors: int = 10,
) -> Dict[str, Any]:
    """
    Estimate intrinsic dimension using local methods.

    Uses correlation dimension and local PCA.
    """
    X = activations.float().cpu().numpy()
    n, d = X.shape

    if n < k_neighbors + 1:
        return {"error": "not enough points for local estimation"}

    # Compute pairwise distances (subsample if large)
    if n > 500:
        idx = np.random.choice(n, 500, replace=False)
        X_sub = X[idx]
    else:
        X_sub = X

    dists = squareform(pdist(X_sub))

    # Local intrinsic dimension via MLE
    # For each point, estimate dimension from k nearest neighbors
    local_dims = []
    for i in range(len(X_sub)):
        # Get k nearest neighbor distances (excluding self)
        d_i = np.sort(dists[i])[1:k_neighbors+1]

        if d_i[-1] > 0 and d_i[0] > 0:
            # MLE estimate: d = k / sum(log(r_k / r_j))
            log_ratios = np.log(d_i[-1] / d_i[:-1])
            if log_ratios.sum() > 0:
                local_dim = (k_neighbors - 1) / log_ratios.sum()
                local_dims.append(local_dim)

    if not local_dims:
        return {"error": "could not estimate local dimensions"}

    local_dims = np.array(local_dims)

    return {
        "local_dim_mean": float(local_dims.mean()),
        "local_dim_std": float(local_dims.std()),
        "local_dim_median": float(np.median(local_dims)),
        "local_dim_min": float(local_dims.min()),
        "local_dim_max": float(local_dims.max()),
    }


def compute_cluster_structure(
    activations: torch.Tensor,
    max_clusters: int = 5,
) -> Dict[str, Any]:
    """
    Analyze cluster structure of the point cloud.
    """
    from sklearn.cluster import KMeans
    from sklearn.metrics import silhouette_score

    X = activations.float().cpu().numpy()
    n, d = X.shape

    if n < 6:
        return {"error": "need at least 6 points"}

    # Try different numbers of clusters
    silhouettes = {}
    inertias = {}

    for k in range(2, min(max_clusters + 1, n // 2)):
        km = KMeans(n_clusters=k, random_state=42, n_init=10)
        labels = km.fit_predict(X)

        try:
            sil = silhouette_score(X, labels)
            silhouettes[k] = float(sil)
        except:
            silhouettes[k] = 0

        inertias[k] = float(km.inertia_)

    # Best k by silhouette
    if silhouettes:
        best_k = max(silhouettes, key=silhouettes.get)
        best_silhouette = silhouettes[best_k]
    else:
        best_k = 1
        best_silhouette = 0

    # Single cluster compactness
    centroid = X.mean(axis=0)
    distances_to_centroid = np.linalg.norm(X - centroid, axis=1)
    compactness = float(distances_to_centroid.mean())

    return {
        "best_k": best_k,
        "best_silhouette": best_silhouette,
        "silhouette_scores": silhouettes,
        "single_cluster_compactness": compactness,
    }


def compute_density_structure(
    activations: torch.Tensor,
    k_neighbors: int = 10,
) -> Dict[str, Any]:
    """
    Analyze density variations in the point cloud.
    """
    X = activations.float().cpu().numpy()
    n, d = X.shape

    if n < k_neighbors + 1:
        return {"error": "not enough points"}

    # Compute local density for each point
    # Density ~ 1 / (distance to k-th neighbor)
    dists = squareform(pdist(X))

    local_densities = []
    for i in range(n):
        kth_dist = np.sort(dists[i])[k_neighbors]
        if kth_dist > 0:
            local_densities.append(1.0 / kth_dist)
        else:
            local_densities.append(float('inf'))

    local_densities = np.array(local_densities)
    finite_densities = local_densities[np.isfinite(local_densities)]

    if len(finite_densities) == 0:
        return {"error": "all densities infinite"}

    # Density uniformity: low std/mean = uniform density
    density_cv = float(finite_densities.std() / (finite_densities.mean() + 1e-8))

    return {
        "density_mean": float(finite_densities.mean()),
        "density_std": float(finite_densities.std()),
        "density_cv": density_cv,  # low = uniform, high = variable
        "density_min": float(finite_densities.min()),
        "density_max": float(finite_densities.max()),
    }


def compute_topology_indicators(
    activations: torch.Tensor,
) -> Dict[str, Any]:
    """
    Basic topological indicators.

    Full persistent homology is expensive, so we use proxies.
    """
    X = activations.float().cpu().numpy()
    n, d = X.shape

    if n < 10:
        return {"error": "need at least 10 points"}

    # Subsample if large
    if n > 300:
        idx = np.random.choice(n, 300, replace=False)
        X = X[idx]
        n = 300

    dists = squareform(pdist(X))

    # Connectivity at different scales
    # At what distance threshold is the point cloud connected?
    sorted_edges = np.sort(dists[np.triu_indices(n, k=1)])

    # Approximate: find distance where graph becomes connected
    # Using minimum spanning tree proxy
    from scipy.sparse.csgraph import minimum_spanning_tree
    from scipy.sparse import csr_matrix

    mst = minimum_spanning_tree(csr_matrix(dists))
    mst_edges = mst.data
    connectivity_radius = float(mst_edges.max()) if len(mst_edges) > 0 else 0

    # Ratio of connectivity radius to mean pairwise distance
    mean_pairwise = float(sorted_edges.mean())
    connectivity_ratio = connectivity_radius / (mean_pairwise + 1e-8)

    return {
        "connectivity_radius": connectivity_radius,
        "mean_pairwise_distance": mean_pairwise,
        "connectivity_ratio": connectivity_ratio,  # low = compact, high = spread/disconnected
    }


def compute_two_cloud_relationship(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
) -> Dict[str, Any]:
    """
    Analyze the geometric relationship between pos and neg clouds.
    """
    pos = pos_activations.float().cpu().numpy()
    neg = neg_activations.float().cpu().numpy()

    n_pos, d = pos.shape
    n_neg = neg.shape[0]

    if n_pos < 3 or n_neg < 3:
        return {"error": "need at least 3 points in each cloud"}

    # Centroids
    pos_centroid = pos.mean(axis=0)
    neg_centroid = neg.mean(axis=0)

    # Distance between centroids
    centroid_distance = float(np.linalg.norm(pos_centroid - neg_centroid))

    # Direction between centroids
    centroid_direction = pos_centroid - neg_centroid
    centroid_dir_norm = np.linalg.norm(centroid_direction)
    if centroid_dir_norm > 1e-8:
        centroid_direction = centroid_direction / centroid_dir_norm

    # Within-class spread
    pos_spread = float(np.linalg.norm(pos - pos_centroid, axis=1).mean())
    neg_spread = float(np.linalg.norm(neg - neg_centroid, axis=1).mean())

    # Separation ratio: distance between centroids / average spread
    avg_spread = (pos_spread + neg_spread) / 2
    separation_ratio = centroid_distance / (avg_spread + 1e-8)

    # Overlap: what fraction of pos is closer to neg_centroid than pos_centroid?
    pos_to_pos_centroid = np.linalg.norm(pos - pos_centroid, axis=1)
    pos_to_neg_centroid = np.linalg.norm(pos - neg_centroid, axis=1)
    pos_overlap = float((pos_to_neg_centroid < pos_to_pos_centroid).mean())

    neg_to_pos_centroid = np.linalg.norm(neg - pos_centroid, axis=1)
    neg_to_neg_centroid = np.linalg.norm(neg - neg_centroid, axis=1)
    neg_overlap = float((neg_to_pos_centroid < neg_to_neg_centroid).mean())

    # Parallelism: do the clouds have similar shape/orientation?
    # Compare principal components
    pos_centered = pos - pos_centroid
    neg_centered = neg - neg_centroid

    try:
        _, _, Vh_pos = np.linalg.svd(pos_centered, full_matrices=False)
        _, _, Vh_neg = np.linalg.svd(neg_centered, full_matrices=False)

        # Alignment of top principal components
        pc1_alignment = float(abs(np.dot(Vh_pos[0], Vh_neg[0])))
        pc2_alignment = float(abs(np.dot(Vh_pos[1], Vh_neg[1]))) if len(Vh_pos) > 1 else 0
    except:
        pc1_alignment = 0
        pc2_alignment = 0

    return {
        # Distance
        "centroid_distance": centroid_distance,
        "separation_ratio": separation_ratio,  # high = well separated

        # Spread
        "pos_spread": pos_spread,
        "neg_spread": neg_spread,

        # Overlap
        "pos_overlap_fraction": pos_overlap,  # low = separated
        "neg_overlap_fraction": neg_overlap,

        # Shape alignment
        "pc1_alignment": pc1_alignment,  # high = parallel manifolds
        "pc2_alignment": pc2_alignment,
    }


def compute_relative_position(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
) -> Dict[str, Any]:
    """
    Analyze relative position: is neg a shifted/rotated version of pos?
    """
    pos = pos_activations.float().cpu().numpy()
    neg = neg_activations.float().cpu().numpy()

    n = min(len(pos), len(neg))
    pos, neg = pos[:n], neg[:n]

    # Centroids
    pos_centroid = pos.mean(axis=0)
    neg_centroid = neg.mean(axis=0)
    shift = pos_centroid - neg_centroid
    shift_norm = np.linalg.norm(shift)

    # If we shift neg by the centroid difference, how close is it to pos?
    neg_shifted = neg + shift

    # After shifting, measure alignment
    residuals = pos - neg_shifted
    residual_norms = np.linalg.norm(residuals, axis=1)

    # Compare to original differences
    orig_diffs = pos - neg
    orig_diff_norms = np.linalg.norm(orig_diffs, axis=1)

    # Shift explains how much of the difference?
    shift_explains = 1 - (residual_norms.mean() / (orig_diff_norms.mean() + 1e-8))

    # Is the relationship a pure translation?
    # If yes, all (pos - neg) vectors should be similar
    diff_normalized = orig_diffs / (orig_diff_norms[:, np.newaxis] + 1e-8)
    mean_diff_dir = diff_normalized.mean(axis=0)
    mean_diff_dir = mean_diff_dir / (np.linalg.norm(mean_diff_dir) + 1e-8)

    translation_consistency = float((diff_normalized @ mean_diff_dir).mean())

    return {
        "shift_vector_norm": float(shift_norm),
        "shift_explains_fraction": float(shift_explains),  # high = pure translation
        "translation_consistency": translation_consistency,  # high = consistent shift
        "residual_after_shift": float(residual_norms.mean()),
    }


def analyze_activation_structure(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
) -> Dict[str, Any]:
    """
    Comprehensive analysis of activation space structure.

    Analyzes:
    - Individual cloud shapes (pos and neg)
    - Relationship between clouds
    - Specific structure tests (cone, sphere, manifold)
    """
    results = {}

    # Individual cloud analysis
    results["pos_shape"] = compute_cloud_shape(pos_activations)
    results["neg_shape"] = compute_cloud_shape(neg_activations)

    # Cone fit
    results["pos_cone"] = compute_cone_fit(pos_activations)
    results["neg_cone"] = compute_cone_fit(neg_activations)

    # Sphere fit
    results["pos_sphere"] = compute_sphere_fit(pos_activations)
    results["neg_sphere"] = compute_sphere_fit(neg_activations)

    # Manifold dimension
    results["pos_manifold"] = compute_manifold_dimension(pos_activations)
    results["neg_manifold"] = compute_manifold_dimension(neg_activations)

    # Cluster structure
    results["pos_clusters"] = compute_cluster_structure(pos_activations)
    results["neg_clusters"] = compute_cluster_structure(neg_activations)

    # Density
    results["pos_density"] = compute_density_structure(pos_activations)
    results["neg_density"] = compute_density_structure(neg_activations)

    # Relationship between clouds
    results["relationship"] = compute_two_cloud_relationship(pos_activations, neg_activations)
    results["relative_position"] = compute_relative_position(pos_activations, neg_activations)

    return results



"""RepScan with automatic concept decomposition and naming."""
from typing import Dict, Any, Optional, List, Tuple
import torch

from .runner import compute_geometry_metrics
from .concept_naming import decompose_and_name_concepts, find_optimal_layer_per_concept
import numpy as np
from .database_loaders import load_activations_from_database, load_pair_texts_from_database

# Re-export database loaders for backwards compatibility
__all__ = [
    "run_repscan_with_concept_naming",
    "extract_pair_texts_from_enriched_pairs",
    "load_activations_from_database",
    "load_pair_texts_from_database",
    "find_optimal_layer_per_concept",
]


def run_repscan_with_concept_naming(
    activations_by_layer: Dict[int, Tuple[torch.Tensor, torch.Tensor]],
    pair_texts: Optional[Dict[int, Dict[str, str]]] = None,
    include_expensive: bool = True,
    generate_visualizations: bool = False,
    llm_model: str = "meta-llama/Llama-3.2-1B-Instruct",
) -> Dict[str, Any]:
    """
    Run RepScan with automatic concept decomposition and LLM naming.

    Args:
        activations_by_layer: Dict mapping layer -> (pos_activations, neg_activations)
        pair_texts: Optional dict mapping pair_id -> {prompt, positive, negative}
        include_expensive: Whether to compute expensive metrics
        generate_visualizations: Whether to generate visualization figures
        llm_model: HuggingFace model for naming

    Returns:
        Dict with:
            - layers: Per-layer metrics
            - best_layer: Layer with highest ICD
            - concept_decomposition: Named concepts with pair assignments
    """
    results = {
        "layers": [],
        "concept_decomposition": None,
    }

    best_layer = None
    best_icd = -1
    best_activations = None

    # Run metrics for each layer
    for layer, (pos, neg) in sorted(activations_by_layer.items()):
        metrics = compute_geometry_metrics(
            pos_activations=pos,
            neg_activations=neg,
            include_expensive=include_expensive,
            generate_visualizations=generate_visualizations,
        )

        layer_result = {
            "layer": layer,
            "n_pos": len(pos),
            "n_neg": len(neg),
            "metrics": metrics,
        }
        results["layers"].append(layer_result)

        # Track best layer
        icd = metrics.get("icd_icd", 0)
        if icd > best_icd:
            best_icd = icd
            best_layer = layer
            best_activations = (pos, neg)

    results["best_layer"] = best_layer
    results["best_layer_icd"] = best_icd

    # Get recommendation from best layer
    for layer_result in results["layers"]:
        if layer_result["layer"] == best_layer:
            results["recommended_method"] = layer_result["metrics"].get("recommended_method")
            results["recommendation_confidence"] = layer_result["metrics"].get("recommendation_confidence")
            break

    # Run concept decomposition on best layer
    if best_activations is not None:
        pos, neg = best_activations
        decomposition = decompose_and_name_concepts(
            pos, neg, pair_texts,
            generate_visualizations=generate_visualizations,
            llm_model=llm_model,
        )

        # Find optimal layer per concept (different concepts may peak at different layers)
        cluster_labels = np.array(decomposition.get("cluster_labels", []))
        n_concepts = decomposition.get("n_concepts", 1)

        if len(cluster_labels) > 0 and n_concepts > 1:
            per_concept_layers = find_optimal_layer_per_concept(
                activations_by_layer, cluster_labels, n_concepts
            )

            # Add optimal layer info to each concept
            for concept in decomposition.get("concepts", []):
                concept_idx = concept["id"] - 1  # 0-indexed
                if concept_idx in per_concept_layers:
                    layer_info = per_concept_layers[concept_idx]
                    concept["optimal_layer"] = layer_info["best_layer"]
                    concept["optimal_layer_accuracy"] = layer_info["best_accuracy"]
                    concept["layer_accuracies"] = layer_info["layer_accuracies"]

            decomposition["per_concept_layers"] = per_concept_layers

        results["concept_decomposition"] = decomposition

    return results


def extract_pair_texts_from_enriched_pairs(enriched_pairs: List[Dict]) -> Dict[int, Dict[str, str]]:
    """
    Extract pair texts from enriched pairs format (wisent CLI output).

    Args:
        enriched_pairs: List of pairs in wisent CLI format

    Returns:
        Dict mapping pair index -> {prompt, positive, negative}
    """
    pair_texts = {}
    for i, pair in enumerate(enriched_pairs):
        pair_texts[i] = {
            "prompt": pair.get("prompt", ""),
            "positive": pair.get("positive_response", {}).get("model_response", ""),
            "negative": pair.get("negative_response", {}).get("model_response", ""),
        }
    return pair_texts



"""
Visualization functions for activation geometry.

All functions return matplotlib figures or data for plotting.
"""

import torch
import numpy as np
from typing import Dict, Any, List, Tuple, Optional
import warnings


def plot_pca_projection(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
    n_components: int = 2,
    title: str = "PCA Projection",
) -> Dict[str, Any]:
    """
    Project activations to 2D or 3D using PCA and return plot data.
    """
    from sklearn.decomposition import PCA

    pos = pos_activations.float().cpu().numpy()
    neg = neg_activations.float().cpu().numpy()

    # Combine for fitting PCA
    X = np.vstack([pos, neg])
    labels = np.array([1] * len(pos) + [0] * len(neg))

    pca = PCA(n_components=n_components)
    X_pca = pca.fit_transform(X)

    pos_pca = X_pca[:len(pos)]
    neg_pca = X_pca[len(pos):]

    return {
        "pos_projected": pos_pca,
        "neg_projected": neg_pca,
        "explained_variance": pca.explained_variance_ratio_.tolist(),
        "components": pca.components_,
        "n_components": n_components,
        "title": title,
    }


def plot_diff_vectors(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
    title: str = "Difference Vectors",
) -> Dict[str, Any]:
    """
    Visualize the difference vectors projected to 2D.
    """
    from sklearn.decomposition import PCA

    pos = pos_activations.float().cpu().numpy()
    neg = neg_activations.float().cpu().numpy()

    n = min(len(pos), len(neg))
    diffs = pos[:n] - neg[:n]

    # Project diffs to 2D
    pca = PCA(n_components=2)
    diffs_2d = pca.fit_transform(diffs)

    # Mean direction
    mean_diff = diffs.mean(axis=0)
    mean_diff_2d = pca.transform(mean_diff.reshape(1, -1))[0]

    return {
        "diffs_projected": diffs_2d,
        "mean_diff_projected": mean_diff_2d,
        "explained_variance": pca.explained_variance_ratio_.tolist(),
        "title": title,
    }


def plot_norm_distribution(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
    title: str = "Norm Distribution",
) -> Dict[str, Any]:
    """
    Plot distribution of activation norms.
    """
    pos = pos_activations.float().cpu().numpy()
    neg = neg_activations.float().cpu().numpy()

    pos_norms = np.linalg.norm(pos, axis=1)
    neg_norms = np.linalg.norm(neg, axis=1)

    return {
        "pos_norms": pos_norms,
        "neg_norms": neg_norms,
        "pos_mean": float(pos_norms.mean()),
        "neg_mean": float(neg_norms.mean()),
        "pos_std": float(pos_norms.std()),
        "neg_std": float(neg_norms.std()),
        "title": title,
    }


def plot_alignment_distribution(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
    title: str = "Alignment Distribution",
) -> Dict[str, Any]:
    """
    Plot distribution of how well each diff aligns with mean diff.
    """
    pos = pos_activations.float().cpu().numpy()
    neg = neg_activations.float().cpu().numpy()

    n = min(len(pos), len(neg))
    diffs = pos[:n] - neg[:n]

    # Mean direction
    mean_diff = diffs.mean(axis=0)
    mean_diff_norm = np.linalg.norm(mean_diff)

    if mean_diff_norm < 1e-8:
        return {"error": "mean diff too small"}

    mean_diff_normalized = mean_diff / mean_diff_norm

    # Per-diff alignment
    diff_norms = np.linalg.norm(diffs, axis=1)
    valid = diff_norms > 1e-8
    alignments = np.zeros(n)
    alignments[valid] = (diffs[valid] / diff_norms[valid, np.newaxis]) @ mean_diff_normalized

    return {
        "alignments": alignments,
        "mean": float(alignments.mean()),
        "std": float(alignments.std()),
        "min": float(alignments.min()),
        "max": float(alignments.max()),
        "title": title,
    }


def plot_eigenvalue_spectrum(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
    n_components: int = 50,
    title: str = "Eigenvalue Spectrum",
) -> Dict[str, Any]:
    """
    Plot eigenvalue spectrum of the difference vectors.
    """
    from sklearn.decomposition import PCA

    pos = pos_activations.float().cpu().numpy()
    neg = neg_activations.float().cpu().numpy()

    n = min(len(pos), len(neg))
    diffs = pos[:n] - neg[:n]

    n_comp = min(n_components, n - 1, diffs.shape[1])
    pca = PCA(n_components=n_comp)
    pca.fit(diffs)

    return {
        "eigenvalues": pca.explained_variance_.tolist(),
        "explained_variance_ratio": pca.explained_variance_ratio_.tolist(),
        "cumulative_variance": np.cumsum(pca.explained_variance_ratio_).tolist(),
        "title": title,
    }


def plot_pairwise_distances(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
    max_samples: int = 200,
    title: str = "Pairwise Distances",
) -> Dict[str, Any]:
    """
    Plot distribution of pairwise distances within and between classes.
    """
    from scipy.spatial.distance import pdist

    pos = pos_activations.float().cpu().numpy()
    neg = neg_activations.float().cpu().numpy()

    # Subsample if needed
    if len(pos) > max_samples:
        idx = np.random.choice(len(pos), max_samples, replace=False)
        pos = pos[idx]
    if len(neg) > max_samples:
        idx = np.random.choice(len(neg), max_samples, replace=False)
        neg = neg[idx]

    # Within-class distances
    pos_dists = pdist(pos)
    neg_dists = pdist(neg)

    # Between-class distances
    from scipy.spatial.distance import cdist
    between_dists = cdist(pos, neg).flatten()

    # Full pairwise distance matrix for visualization
    all_data = np.vstack([pos, neg])
    distance_matrix = cdist(all_data, all_data)

    return {
        "pos_within": pos_dists,
        "neg_within": neg_dists,
        "between": between_dists,
        "pos_within_mean": float(pos_dists.mean()),
        "neg_within_mean": float(neg_dists.mean()),
        "between_mean": float(between_dists.mean()),
        "distance_matrix": distance_matrix,
        "n_pos": len(pos),
        "title": title,
    }


def plot_tsne_projection(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
    perplexity: int = 30,
    title: str = "t-SNE Projection",
) -> Dict[str, Any]:
    """
    Project activations using t-SNE.
    """
    try:
        from sklearn.manifold import TSNE
    except ImportError:
        return {"error": "sklearn not available"}

    pos = pos_activations.float().cpu().numpy()
    neg = neg_activations.float().cpu().numpy()

    X = np.vstack([pos, neg])

    # Adjust perplexity if needed
    n_samples = len(X)
    perplexity = min(perplexity, n_samples - 1)

    if n_samples < 5:
        return {"error": "not enough samples for t-SNE"}

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        tsne = TSNE(n_components=2, perplexity=perplexity, random_state=42)
        X_tsne = tsne.fit_transform(X)

    pos_tsne = X_tsne[:len(pos)]
    neg_tsne = X_tsne[len(pos):]

    return {
        "pos_projected": pos_tsne,
        "neg_projected": neg_tsne,
        "title": title,
    }


def plot_umap_projection(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
    n_neighbors: int = 15,
    min_dist: float = 0.1,
    title: str = "UMAP Projection",
) -> Dict[str, Any]:
    """
    Project activations using UMAP.
    """
    try:
        import umap
    except ImportError:
        return {"error": "umap not available"}

    pos = pos_activations.float().cpu().numpy()
    neg = neg_activations.float().cpu().numpy()

    X = np.vstack([pos, neg])

    n_samples = len(X)
    n_neighbors = min(n_neighbors, n_samples - 1)

    if n_samples < 5:
        return {"error": "not enough samples for UMAP"}

    reducer = umap.UMAP(n_components=2, n_neighbors=n_neighbors, min_dist=min_dist, random_state=42)
    X_umap = reducer.fit_transform(X)

    pos_umap = X_umap[:len(pos)]
    neg_umap = X_umap[len(pos):]

    return {
        "pos_projected": pos_umap,
        "neg_projected": neg_umap,
        "title": title,
    }


def plot_pacmap_projection(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
    n_neighbors: int = 10,
    num_iters: int = 50,
    pca_dims: int = 30,
    title: str = "PaCMAP Projection",
) -> Dict[str, Any]:
    """
    Project activations using PaCMAP (alternative implementation).

    Uses pacmap_alt which avoids numba threading issues on macOS.
    """
    from .pacmap_alt import plot_pacmap_alt

    return plot_pacmap_alt(
        pos_activations,
        neg_activations,
        n_neighbors=n_neighbors,
        num_iters=num_iters,
        title=title,
    )


def plot_cone_visualization(
    activations: torch.Tensor,
    title: str = "Cone Structure",
) -> Dict[str, Any]:
    """
    Visualize how cone-like the activations are.

    Projects to 2D and shows angular distribution from mean direction.
    """
    X = activations.float().cpu().numpy()

    # Normalize
    norms = np.linalg.norm(X, axis=1, keepdims=True)
    valid = norms.squeeze() > 1e-8
    X_norm = X[valid] / norms[valid]

    # Mean direction
    mean_dir = X_norm.mean(axis=0)
    mean_dir = mean_dir / (np.linalg.norm(mean_dir) + 1e-8)

    # Angles from mean
    cos_angles = X_norm @ mean_dir
    angles = np.degrees(np.arccos(np.clip(cos_angles, -1, 1)))

    # Project to 2D: mean direction and orthogonal
    from sklearn.decomposition import PCA
    pca = PCA(n_components=2)
    X_2d = pca.fit_transform(X_norm)

    return {
        "angles_from_mean": angles,
        "projected_2d": X_2d,
        "mean_angle": float(angles.mean()),
        "max_angle": float(angles.max()),
        "title": title,
    }


def plot_layer_comparison(
    activations_by_layer: Dict[int, Tuple[torch.Tensor, torch.Tensor]],
    metric_fn,
    metric_name: str = "metric",
    title: str = "Metric Across Layers",
) -> Dict[str, Any]:
    """
    Plot a metric across different layers.
    """
    layers = sorted(activations_by_layer.keys())
    values = []

    for layer in layers:
        pos, neg = activations_by_layer[layer]
        try:
            result = metric_fn(pos, neg)
            if isinstance(result, dict):
                # Try to extract a scalar
                val = result.get(metric_name, result.get("mean", 0))
            else:
                val = result
            values.append(float(val))
        except:
            values.append(0)

    return {
        "layers": layers,
        "values": values,
        "metric_name": metric_name,
        "title": title,
    }


def create_summary_figure(
    pos_activations: torch.Tensor,
    neg_activations: torch.Tensor,
    metrics: Dict[str, Any] = None,
    include_pacmap: bool = True,
) -> str:
    """
    Create a multi-panel summary figure and return as base64.

    Creates a 3x3 grid with all 9 visualizations:
    Row 1: PCA, t-SNE, UMAP
    Row 2: PaCMAP, Diff Vectors, Alignment Distribution
    Row 3: Eigenvalue Spectrum, Norm Distribution, Pairwise Distances

    Returns:
        Base64-encoded PNG string
    """
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import io
    import base64

    fig, axes = plt.subplots(3, 3, figsize=(18, 18))

    def plot_projection(ax, data, title):
        """Helper to plot a 2D projection."""
        if "error" in data:
            ax.text(0.5, 0.5, f"Error: {data['error']}", ha='center', va='center', transform=ax.transAxes)
        else:
            pos = data["pos_projected"]
            neg = data["neg_projected"]
            ax.scatter(pos[:, 0], pos[:, 1], c='blue', alpha=0.6, label='Pos', s=20)
            ax.scatter(neg[:, 0], neg[:, 1], c='red', alpha=0.6, label='Neg', s=20)
            ax.legend(loc='upper right', fontsize=8)
        ax.set_title(title, fontsize=11)

    # Row 1: Dimensionality reduction projections
    # 1. PCA projection
    pca_data = plot_pca_projection(pos_activations, neg_activations)
    plot_projection(axes[0, 0], pca_data, "PCA Projection")

    # 2. t-SNE projection
    tsne_data = plot_tsne_projection(pos_activations, neg_activations)
    plot_projection(axes[0, 1], tsne_data, "t-SNE Projection")

    # 3. UMAP projection
    umap_data = plot_umap_projection(pos_activations, neg_activations)
    plot_projection(axes[0, 2], umap_data, "UMAP Projection")

    # Row 2: More projections and distributions
    # 4. PaCMAP projection
    if include_pacmap:
        try:
            from .pacmap_alt import plot_pacmap_alt
            pacmap_data = plot_pacmap_alt(pos_activations, neg_activations)
        except Exception as e:
            pacmap_data = {"error": str(e)}
    else:
        pacmap_data = {"error": "PaCMAP disabled"}
    plot_projection(axes[1, 0], pacmap_data, "PaCMAP Projection")

    # 5. Diff vectors
    diff_data = plot_diff_vectors(pos_activations, neg_activations)
    ax = axes[1, 1]
    if "error" not in diff_data:
        diffs = diff_data["diffs_projected"]
        mean = diff_data["mean_diff_projected"]
        ax.scatter(diffs[:, 0], diffs[:, 1], c='green', alpha=0.6, s=20)
        scale = max(abs(mean[0]), abs(mean[1]), 0.1) * 0.8
        ax.arrow(0, 0, mean[0]*0.8, mean[1]*0.8, head_width=scale*0.1, head_length=scale*0.05, fc='black', ec='black')
    ax.set_title("Diff Vectors (Mean Direction)", fontsize=11)

    # 6. Alignment distribution
    align_data = plot_alignment_distribution(pos_activations, neg_activations)
    ax = axes[1, 2]
    if "error" not in align_data:
        ax.hist(align_data["alignments"], bins=20, color='green', alpha=0.7)
        ax.axvline(align_data["mean"], color='black', linestyle='--', linewidth=2)
        ax.set_title(f"Alignment (mean={align_data['mean']:.2f})", fontsize=11)
    else:
        ax.text(0.5, 0.5, "Error", ha='center', va='center', transform=ax.transAxes)
        ax.set_title("Alignment Distribution", fontsize=11)
    ax.set_xlabel("Cosine Alignment")
    ax.set_ylabel("Count")

    # Row 3: Spectral and distance analysis
    # 7. Eigenvalue spectrum
    eigen_data = plot_eigenvalue_spectrum(pos_activations, neg_activations)
    ax = axes[2, 0]
    n_show = min(20, len(eigen_data["explained_variance_ratio"]))
    ax.bar(range(n_show), eigen_data["explained_variance_ratio"][:n_show], color='purple', alpha=0.7)
    ax.set_title("Eigenvalue Spectrum", fontsize=11)
    ax.set_xlabel("Principal Component")
    ax.set_ylabel("Variance Ratio")

    # 8. Norm distribution
    norm_data = plot_norm_distribution(pos_activations, neg_activations)
    ax = axes[2, 1]
    ax.hist(norm_data["pos_norms"], bins=20, alpha=0.5, label='Pos', color='blue')
    ax.hist(norm_data["neg_norms"], bins=20, alpha=0.5, label='Neg', color='red')
    ax.legend(fontsize=8)
    ax.set_title("Norm Distribution", fontsize=11)
    ax.set_xlabel("L2 Norm")
    ax.set_ylabel("Count")

    # 9. Pairwise distances
    dist_data = plot_pairwise_distances(pos_activations, neg_activations)
    ax = axes[2, 2]
    if "error" not in dist_data:
        im = ax.imshow(dist_data["distance_matrix"], cmap='viridis', aspect='auto')
        n_pos = dist_data.get("n_pos", len(pos_activations))
        ax.axhline(n_pos - 0.5, color='white', linestyle='--', linewidth=1)
        ax.axvline(n_pos - 0.5, color='white', linestyle='--', linewidth=1)
        plt.colorbar(im, ax=ax, fraction=0.046, pad=0.04)
    ax.set_title("Pairwise Distances", fontsize=11)
    ax.set_xlabel("Sample Index")
    ax.set_ylabel("Sample Index")

    # Add overall title with key metrics if available
    if metrics:
        # Include concept title if available
        title_prefix = metrics.get('concept_title', 'Layer Analysis')
        suptitle = f"{title_prefix} | Linear Probe: {metrics.get('linear_probe_accuracy', 0):.2f} | ICD: {metrics.get('icd_icd', 0):.1f} | Rec: {metrics.get('recommended_method', 'N/A')}"
        fig.suptitle(suptitle, fontsize=14, y=1.01)

    plt.tight_layout()

    # Convert to base64
    buf = io.BytesIO()
    fig.savefig(buf, format='png', dpi=120, bbox_inches='tight')
    buf.seek(0)
    img_base64 = base64.b64encode(buf.read()).decode('utf-8')
    plt.close(fig)

    return img_base64


def render_matplotlib_figure(
    plot_data: Dict[str, Any],
    plot_type: str = None,
    figsize: Tuple[int, int] = (8, 6),
    return_base64: bool = True,
) -> str:
    """
    Render plot data to a matplotlib figure and return as base64.

    Args:
        plot_data: Output from one of the plot_* functions
        plot_type: One of 'pca', 'diff_vectors', 'norms', 'alignments',
                   'eigenvalues', 'distances', 'tsne', 'umap', 'cone'.
                   If None, auto-detect from plot_data.
        figsize: Figure size
        return_base64: If True, return base64 string; if False, return figure

    Returns:
        Base64-encoded PNG string (or figure if return_base64=False)
    """
    import matplotlib
    matplotlib.use('Agg')  # Non-interactive backend
    import matplotlib.pyplot as plt
    import io
    import base64

    # Auto-detect plot type from data
    if plot_type is None:
        if "pos_projected" in plot_data and "neg_projected" in plot_data:
            plot_type = "pca"
        elif "diffs_projected" in plot_data:
            plot_type = "diff_vectors"
        elif "pos_norms" in plot_data:
            plot_type = "norms"
        elif "alignments" in plot_data:
            plot_type = "alignments"
        elif "explained_variance_ratio" in plot_data:
            plot_type = "eigenvalues"
        elif "pos_within" in plot_data:
            plot_type = "distances"
        elif "angles_from_mean" in plot_data:
            plot_type = "cone"
        else:
            plot_type = "unknown"

    fig, ax = plt.subplots(figsize=figsize)

    if "error" in plot_data:
        ax.text(0.5, 0.5, f"Error: {plot_data['error']}", ha='center', va='center')
    else:
        title = plot_data.get("title", plot_type)

        if plot_type in ['pca', 'tsne', 'umap']:
            pos = plot_data["pos_projected"]
            neg = plot_data["neg_projected"]
            ax.scatter(pos[:, 0], pos[:, 1], c='blue', alpha=0.6, label='Positive')
            ax.scatter(neg[:, 0], neg[:, 1], c='red', alpha=0.6, label='Negative')
            ax.legend()
            ax.set_xlabel("Component 1")
            ax.set_ylabel("Component 2")

        elif plot_type == 'diff_vectors':
            diffs = plot_data["diffs_projected"]
            mean = plot_data["mean_diff_projected"]
            ax.scatter(diffs[:, 0], diffs[:, 1], c='green', alpha=0.6)
            ax.arrow(0, 0, mean[0], mean[1], head_width=0.1, head_length=0.05, fc='black', ec='black')
            ax.set_xlabel("PC1")
            ax.set_ylabel("PC2")

        elif plot_type == 'norms':
            ax.hist(plot_data["pos_norms"], bins=20, alpha=0.5, label='Positive', color='blue')
            ax.hist(plot_data["neg_norms"], bins=20, alpha=0.5, label='Negative', color='red')
            ax.legend()
            ax.set_xlabel("Norm")
            ax.set_ylabel("Count")

        elif plot_type == 'alignments':
            ax.hist(plot_data["alignments"], bins=20, color='green', alpha=0.7)
            ax.axvline(plot_data["mean"], color='black', linestyle='--', label=f'Mean: {plot_data["mean"]:.2f}')
            ax.legend()
            ax.set_xlabel("Alignment with mean direction")
            ax.set_ylabel("Count")

        elif plot_type == 'eigenvalues':
            ax.bar(range(len(plot_data["explained_variance_ratio"])), plot_data["explained_variance_ratio"])
            ax.set_xlabel("Principal Component")
            ax.set_ylabel("Explained Variance Ratio")

        elif plot_type == 'distances':
            ax.hist(plot_data["pos_within"], bins=30, alpha=0.5, label='Within Pos', color='blue')
            ax.hist(plot_data["neg_within"], bins=30, alpha=0.5, label='Within Neg', color='red')
            ax.hist(plot_data["between"], bins=30, alpha=0.5, label='Between', color='green')
            ax.legend()
            ax.set_xlabel("Distance")
            ax.set_ylabel("Count")

        elif plot_type == 'cone':
            ax.hist(plot_data["angles_from_mean"], bins=20, color='purple', alpha=0.7)
            ax.axvline(plot_data["mean_angle"], color='black', linestyle='--',
                       label=f'Mean: {plot_data["mean_angle"]:.1f}°')
            ax.legend()
            ax.set_xlabel("Angle from mean direction (degrees)")
            ax.set_ylabel("Count")

        ax.set_title(title)

    plt.tight_layout()

    if not return_base64:
        return fig

    # Convert to base64
    buf = io.BytesIO()
    fig.savefig(buf, format='png', dpi=100, bbox_inches='tight')
    buf.seek(0)
    img_base64 = base64.b64encode(buf.read()).decode('utf-8')
    plt.close(fig)

    return img_base64



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from wisent.core.contrastive_pairs.core.pair import ContrastivePair
from wisent.core.contrastive_pairs.core.response import NegativeResponse, PositiveResponse
from wisent.core.contrastive_pairs.lm_eval_pairs.atoms import LMEvalBenchmarkExtractor
from wisent.core.cli_logger import setup_logger, bind

if TYPE_CHECKING:
    from lm_eval.api.task import ConfigurableTask


__all__ = ["AssinExtractor"]
_LOG = setup_logger(__name__)

task_names = ("assin_entailment", "assin_paraphrase")

class AssinExtractor(LMEvalBenchmarkExtractor):
    """Extractor for Assin benchmark - entailment task."""


    evaluator_name = "log_likelihoods"
    def extract_contrastive_pairs(
        self,
        lm_eval_task_data: ConfigurableTask,
        limit: int | None = None,
        preferred_doc: str | None = None,
    ) -> list[ContrastivePair]:
        log = bind(_LOG, task=getattr(lm_eval_task_data, "NAME", "unknown"))
        max_items = self._normalize_limit(limit)
        docs = self.load_docs(lm_eval_task_data, max_items, preferred_doc=preferred_doc)
        pairs: list[ContrastivePair] = []
        log.info("Extracting contrastive pairs", extra={"doc_count": len(docs)})

        for doc in docs:
            pair = self._extract_pair_from_doc(doc)
            if pair is not None:
                pairs.append(pair)
                if max_items is not None and len(pairs) >= max_items:
                    break

        if not pairs:
            task_name = getattr(lm_eval_task_data, "NAME", type(lm_eval_task_data).__name__)
            log.warning("No valid pairs extracted", extra={"task": task_name})

        return pairs

    def _extract_pair_from_doc(self, doc: dict[str, Any]) -> ContrastivePair | None:
        log = bind(_LOG, doc_id=doc.get("id", "unknown"))

        try:
            # Assin entailment format has premise, hypothesis, and entailment_judgment
            premise = doc.get("premise", "").strip()
            hypothesis = doc.get("hypothesis", "").strip()
            entailment_judgment = doc.get("entailment_judgment")

            if not premise or not hypothesis or entailment_judgment is None:
                log.debug("Skipping doc due to missing premise/hypothesis/entailment", extra={"doc": doc})
                return None

            # Build the two choices following lm-eval format:
            # Choice 0: "premise, certo? Também, hypothesis"  (no entailment)
            # Choice 1: "premise, certo? Sim, hypothesis"     (entailment)
            choice_0 = f"{premise}, certo? Também, {hypothesis}"
            choice_1 = f"{premise}, certo? Sim, {hypothesis}"

            # entailment_judgment: 0 = no entailment (choice 0), 1 = entailment (choice 1)
            if entailment_judgment == 0:
                correct = choice_0
                incorrect = choice_1
            elif entailment_judgment == 1:
                correct = choice_1
                incorrect = choice_0
            else:
                log.debug("Invalid entailment_judgment value", extra={"entailment_judgment": entailment_judgment})
                return None

            # Use premise as prompt (though the choices contain full text for log-likelihood)
            prompt = premise
            metadata = {"label": "assin_entailment"}

            return self._build_pair(
                question=prompt,
                correct=correct,
                incorrect=incorrect,
                metadata=metadata,
            )

        except Exception as exc:
            log.error("Error extracting pair from doc", exc_info=exc, extra={"doc": doc})
            return None

    @staticmethod
    def _build_pair(
        question: str,
        correct: str,
        incorrect: str,
        metadata: dict[str, Any] | None = None,
    ) -> ContrastivePair:
        positive_response = PositiveResponse(model_response=correct)
        negative_response = NegativeResponse(model_response=incorrect)
        return ContrastivePair(prompt=question, positive_response=positive_response, negative_response=negative_response, label=metadata.get("label"))
